# -*- coding: utf-8 -*-

"""Collection of physical constants and conversion factors.

The magnitudes of the defined constants are taken from
:mod:`typhon.constants`.

This module adds units defined with pint's UnitRegistry..

Physical constants
==================

============================  ============================
``g``                         Earth standard gravity
``h``                         Planck constant
``k``                         Boltzmann constant
``c``                         Speed of light
``N_A``                       Avogadro constant
``R``                         Universal gas constant
``molar_mass_dry_air``        Molar mass for dry air
``molar_mass_water``          Molar mass for water vapor
``gas_constant_dry_air``      Gas constant for dry air
``gas_constant_water_vapor``  Gas constant for water vapor
============================  ============================

Mathematical constants
======================

==========  ============
``golden``  Golden ratio
==========  ============

SI prefixes
===========

=========  ================
``yotta``  :math:`10^{24}`
``zetta``  :math:`10^{21}`
``exa``    :math:`10^{18}`
``peta``   :math:`10^{15}`
``tera``   :math:`10^{12}`
``giga``   :math:`10^{9}`
``mega``   :math:`10^{6}`
``kilo``   :math:`10^{3}`
``hecto``  :math:`10^{2}`
``deka``   :math:`10^{1}`
``deci``   :math:`10^{-1}`
``centi``  :math:`10^{-2}`
``milli``  :math:`10^{-3}`
``micro``  :math:`10^{-6}`
``nano``   :math:`10^{-9}`
``pico``   :math:`10^{-12}`
``femto``  :math:`10^{-15}`
``atto``   :math:`10^{-18}`
``zepto``  :math:`10^{-21}`
=========  ================

Non-SI ratios
=============

=======  =====================================
``ppm``  :math:`10^{-6}` `parts per million`
``ppb``  :math:`10^{-9}` `parts per billion`
``ppt``  :math:`10^{-12}` `parts per trillion`
=======  =====================================

Binary prefixes
===============

=================  ==============
``kibi``, ``KiB``  :math:`2^{10}`
``mebi``, ``MiB``  :math:`2^{20}`
``gibi``           :math:`2^{30}`
``tebi``           :math:`2^{40}`
``pebi``           :math:`2^{50}`
``exbi``           :math:`2^{60}`
``zebi``           :math:`2^{70}`
``yobi``           :math:`2^{80}`
=================  ==============

=================  ==============
``KB``             :math:`10^3`
``MB``             :math:`10^6`
=================  ==============

Earth characteristics
=====================

================  ===================
``earth_mass``    Earth mass
``earth_radius``  Earth radius
``atm``           Standard atmosphere
================  ===================

"""
import numpy as np

from typhon import constants
from typhon.physics.units.common import ureg


# Physcial constants
g = earth_standard_gravity = constants.g * ureg('m / s**2')
h = planck = constants.planck * ureg.joule
k = boltzmann = constants.boltzmann * ureg('J / K')
c = speed_of_light = constants.speed_of_light * ureg('m / s')
N_A = avogadro = N = constants.avogadro * ureg('1 / mol')
R = gas_constant = constants.gas_constant * ureg('J * mol**-1 * K**-1')
molar_mass_dry_air = 28.9645e-3 * ureg('kg / mol')
molar_mass_water = 18.01528e-3 * ureg('kg / mol')
gas_constant_dry_air = R / molar_mass_dry_air  # J K^-1 kg^-1
gas_constant_water_vapor = R / molar_mass_water  # J K^-1 kg^-1

# Mathematical constants
golden = golden_ratio = (1 + np.sqrt(5)) / 2

# SI prefixes
yotta = 1e24
zetta = 1e21
exa = 1e18
peta = 1e15
tera = 1e12
giga = 1e9
mega = 1e6
kilo = 1e3
hecto = 1e2
deka = 1e1
deci = 1e-1
centi = 1e-2
milli = 1e-3
micro = 1e-6
nano = 1e-9
pico = 1e-12
femto = 1e-15
atto = 1e-18
zepto = 1e-21

# Non-SI ratios
ppm = 1e-6  # parts per million
ppb = 1e-9  # parts per billion
ppt = 1e-12  # parts per trillion

# Binary prefixes
kibi = KiB = 2**10
mebi = MiB = 2**20
gibi = 2**30
tebi = 2**40
pebi = 2**50
exbi = 2**60
zebi = 2**70
yobi = 2**80

KB = 10**3
MB = 10**6

# Earth characteristics
earth_mass = constants.earth_mass * ureg.kg
earth_radius = constants.earth_radius * ureg.m
atm = atmosphere = constants.atm * ureg.pascal



"""Typhon is a collection of tools for atmospheric research.

Typhon provides:

- reading and writing routines for ARTS XML files
- an API to run and access ARTS through Python
- conversion routines for various physical quantities
- a tool kit to collocate different data sets (e.g. satellite, ship, ...)
- different retrievals (e.g. QRNN, SPARE-ICE, ...)
- a subset of the cmocean color maps
- functions to calculate and analyse cloud masks
- various plotting utility functions
- functions for geodetic and geographical calculations
- interface to the SRTM30 global elevation model
- functions for cloudmask statistics
- and much more...

Further information on ARTS can be found on http://www.radiativetransfer.org/.
"""

import logging
import subprocess

from setuptools import setup, find_packages
from codecs import open
from os.path import dirname, join

import builtins

builtins.__TYPHON_SETUP__ = True
DOCLINES = (__doc__ or "").split("\n")

version = open(join(dirname(__file__), "typhon", "VERSION")).read().strip()

if "dev" in version:
    try:
        cp = subprocess.run(
            ["git", "describe", "--tags"], stdout=subprocess.PIPE, check=True
        )
    except subprocess.CalledProcessError:
        logging.warning(
            "Warning: could not determine version from git, "
            "using version from source"
        )
    else:
        so = cp.stdout
        version = (
            so.strip()
            .decode("ascii")
            .lstrip("v")
            .replace("-", "+dev", 1)
            .replace("-", ".")
        )

__version__ = version

setup(
    name="typhon",
    author="The Typhon developers",
    author_email="typhon.mi@lists.uni-hamburg.de",
    version=__version__,
    url="https://github.com/atmtools/typhon",
    download_url="https://github.com/atmtools/typhon/tarball/v" + __version__,
    packages=find_packages(),
    license="MIT",
    description=DOCLINES[0],
    long_description="\n".join(DOCLINES[2:]),
    classifiers=[
        # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires="~=3.6",
    include_package_data=True,
    install_requires=[
        "docutils",
        "imageio",
        "matplotlib>=1.4",
        "netCDF4>=1.1.1",
        "numba",
        "numexpr",
        "numpy>=1.13",
        "pandas",
        "scikit-image",
        "scikit-learn",
        "scipy>=0.15.1",
        "setuptools>=0.7.2",
        "xarray>=0.10.2",
    ],
    extras_require={
        "docs": ["cartopy", "pint", "sphinx_rtd_theme"],
        "tests": ["pytest", "pint", "gdal"],
    },
)



# -*- coding: utf-8 -*-

"""Common functions for typhon.math.
"""
import numpy as np
import functools

__all__ = [
    'cantor_pairing',
    'integrate_column',
    'interpolate_halflevels',
    'sum_digits',
    'nlogspace',
    'promote_maximally',
    'calculate_precisely',
    'squeezable_logspace',
    'multiple_logical',
]


def cantor_pairing(a, b, ):
    """Create an unique number from two natural numbers

    For more information about the Cantor pairing function, have a look at:
    https://en.wikipedia.org/wiki/Pairing_function

    This create an unique number from two natural numbers according to
    .. math::
        \pi (a,b):={\frac{1}{2}}(a+b)(a+b+1)+b.

    Args:
        a: A numpy.array with natural numbers, i.e. unsigned integer.
        b: A numpy.array with natural numbers, i.e. unsigned integer.

    Returns:
        A numpy.array with the unique values.
    """

    a_b_sum = a + b
    return (0.5 * a_b_sum * (a_b_sum+1) + b).astype("int")


def integrate_column(y, x=None, axis=0):
    """Integrate array along an arbitrary axis.

    Note:
        This function is just a wrapper for :func:`numpy.trapz`.

    Parameters:
        y (ndarray): Data array.
        x (ndarray): Coordinate array.
        axis (int): Axis to integrate along for multidimensional input.

    Returns:
        float or ndarray: Column integral.

    Examples:
        >>> import numpy as np
        >>> x = np.linspace(0, 1, 5)
        >>> y = np.arange(5)
        >>> integrate_column(y)
        8.0
        >>> integrate_column(y, x)
        2.0
    """
    return np.trapz(y, x, axis=axis)


def interpolate_halflevels(x, axis=0):
    """Returns the linear inteprolated halflevels for given array.

    Parameters:
        x (ndarray): Data array.
        axis (int): Axis to interpolate along.

    Returns:
        ndarray: Values at halflevels.

    Examples:
        >>> interpolate_halflevels([0, 1, 2, 4])
        array([ 0.5,  1.5,  3. ])
    """
    return (np.take(x, range(1, np.shape(x)[axis]), axis=axis) +
            np.take(x, range(0, np.shape(x)[axis] - 1), axis=axis)) / 2


def sum_digits(n):
    """Calculate the sum of digits.

    Parameters:
       n (int): Number.

    Returns:
        int: Sum of digitis of n.

    Examples:
        >>> sum_digits(42)
        6

    """
    s = 0
    while n:
        s += n % 10
        n //= 10

    return s


def nlogspace(start, stop, num=50):
    """Creates a vector with equally logarithmic spacing.

    Creates a vector with length num, equally logarithmically
    spaced between the given end values.

    Parameters:
        start (int): The starting value of the sequence.
        stop (int): The end value of the sequence,
            unless `endpoint` is set to False.
        num (int): Number of samples to generate.
            Default is 50. Must be non-negative.

    Returns: ndarray.

    Examples:
        >>> nlogspace(10, 1000, 3)
        array([  10.,  100., 1000.])
    """
    return np.exp(np.linspace(np.log(start), np.log(stop), num))

# Any commits made to this module between 2015-05-01 and 2017-03-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822.  That specifically applies to the functions
# promote_maximally and calculate_precisely.
# 
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

def promote_maximally(x):
    """Return copy of x with high precision dtype.

    Converts input of 'f2', 'f4', or 'f8' to 'f8'.  Please don't pass f16.
    f16 is misleading and naughty.

    Converts input of 'u1', 'u2', 'u4', 'u8' to 'u8'.

    Converts input of 'i1', 'i2', 'i4', 'i8' to 'i8'.

    Naturally, this copies the data and increases memory usage.

    Anything else is returned unchanged.

    If you input a pint quantity you will get back a pint quantity.

    Experimental function.
    """
    try:
        q = x.m
        u = x.u
    except AttributeError:  # not a pint quantity
        q = x
        u = None
    try:
        kind = q.dtype.kind
    except AttributeError:  # not from numpy
        return q
    if kind in "fiu":
        newx = q.astype(kind + "8")
        return newx*u if u else newx
    else:
        return x


def calculate_precisely(f):
    """Raise all arguments to their highest numpy precision.

    This decorator copies any floats to f8, ints to i8, preserving masked
    arrays and/or ureg units.

    Currently only works for simple dtypes of float, int, or uint.

    This makes a copy.  Therefore, it is memory-intensive and it does not
    work if function has need to change values in-place.

    Experimental function.

    See also: https://github.com/numpy/numpy/issues/593
    """

    # NB: this decorator supports pint but does not depend on it
    @functools.wraps(f)
    def inner(*args, **kwargs):
        newargs = []
        for arg in args:
            try:
                newargs.append(promote_maximally(arg))
            except AttributeError:
                newargs.append(arg)
        newkwargs = {}
        for (k, v) in kwargs.items():
            try:
                newkwargs[k] = promote_maximally(v)
            except AttributeError:
                newkwargs[k] = v
        return f(*newargs, **newkwargs)
    return inner


def squeezable_logspace(start, stop, num=50, squeeze=1., fixpoint=0.):
    """Create a logarithmic grid that is squeezable around a fixpoint.

    Parameters:
        start (float): The starting value of the sequence.
        stop (float): The end value of the sequence.
        num (int): Number of sample to generate (Default is 50).
        squeeze (float): Factor with which the first stepwidth is
            squeezed in logspace. Has to be between  ``(0, 2)``.
            Values smaller than one compress the gridpoints,
            while values greater than 1 strecht the spacing.
            The default is ``1`` (do not squeeze.)
        fixpoint (float): Relative fixpoint for squeezing the grid.
            Has to be between ``[0, 1]``. The  default is ``0`` (bottom).

    Examples:
        Constructing an unsqueezed grid in logspace.

        >>> squeezable_logspace(1, 100, num=5)
        array([1., 3.16227766, 10., 31.6227766, 100.])

        Constructing a grid that is squeezed at the start.

        >>> squeezable_logspace(1, 100, num=5, squeeze=0.5)
        array([1., 1.77827941, 4.64158883, 17.7827941, 100.])

        Constructing a grid that is squeezed in the middle.

        >>> squeezable_logspace(1, 100, num=5, squeeze=0.5, fixpoint=0.5)
        array([1., 5.62341325, 10., 17.7827941, 100.])

        Visualization of different fixpoint and squeeze factor combinations.

        .. plot::
            :include-source:

            import itertools

            from typhon.plots import profile_p_log
            from typhon.math import squeezable_logspace


            fixpoints = [0, 0.7]
            squeezefacotrs = [0.5, 1.5]
            combinations = itertools.product(fixpoints, squeezefacotrs)

            fig, axes = plt.subplots(len(fixpoints), len(squeezefacotrs),
                                    sharex=True, sharey=True)
            for ax, (fp, s) in zip(axes.flat, combinations):
                p = squeezable_logspace(1000e2, 0.01e2, 20,
                                        fixpoint=fp, squeeze=s)
                profile_p_log(p, np.ones(p.size),
                            marker='.', linestyle='none', ax=ax)
                ax.set_title('fixpoint={}, squeeze={}'.format(fp, s),
                            size='x-small')

    Returns:
        ndarray: (Squeezed) logarithmic grid.
    """
    # The squeeze factor has to be between 0 and 2. Otherwise, the order
    # of groidpoints is arbitrarily swapped as the scaled stepsizes
    # become negative.
    if squeeze <= 0 or squeeze >= 2:
        raise ValueError(
            'Squeeze factor has to be in the open interval (0, 2).'
        )

    # The fixpoint has to be between 0 and 1. It is used as a relative index
    # within the grid, values exceeding the limits result in an IndexError.
    if fixpoint < 0 or fixpoint > 1:
        raise ValueError(
            'The fixpoint has to be in the closed interval [0, 1].'
        )

    # Convert the (relative) fixpoint into an index dependent on gridsize.
    fixpoint_index = int(fixpoint * (num - 1))

    # Create a gridpoints with constant spacing in log-scale.
    samples = np.linspace(np.log(start), np.log(stop), num)

    # Select the bottom part of the grid. The gridpoint is included in the
    # bottom and top part to ensure right step widths.
    bottom = samples[:fixpoint_index + 1]

    # Calculate the stepsizes between each gridpoint.
    steps = np.diff(bottom)

    # Re-adjust the stepwidth according to given squeeze factor.
    # The squeeze factor is linearly changing in logspace.
    steps *= np.linspace(2 - squeeze, squeeze, steps.size)

    # Reconstruct the actual gridpoints by adding the first grid value and
    # the cumulative sum of the stepwidths.
    bottom = bottom[0] + np.cumsum(np.append([0], steps))

    # Re-adjust the top part as stated above.
    # **The order of squeezing is inverted!**
    top = samples[fixpoint_index:]
    steps = np.diff(top)
    steps *= np.linspace(squeeze, 2 - squeeze, steps.size)
    top = top[0] + np.cumsum(np.append([0], steps))

    # Combine the bottom and top parts to the final grid. Drop the fixpoint
    # in the bottom part to avoid duplicates.
    return np.exp(np.append(bottom[:-1], top))


def multiple_logical(*args, func=np.logical_or):
    """Apply logical function on multiple arguments.

    Parameters:
        *args (ndarray[bool]): Arrays to be combined.
        func (callable): Logical function, default :func:`numpy.logical_or`.

    Returns:
        ndarray[bool]: Combined boolean array.
    """
    mask = args[0]
    for arg in args[1:]:
        mask = func(mask, arg)

    return mask



# -*- coding: utf-8 -*-
"""Statistical functions for binary cloud masks. """
import numpy as np
import scipy as sc

from skimage import measure
from scipy.spatial.distance import pdist


__all__ = [
    'get_cloudproperties',
    'neighbor_distance',
    'iorg',
    'scai',
    'cloudfraction',
]


def get_cloudproperties(cloudmask, connectivity=1):
    """Calculate basic cloud properties from binary cloudmask.

    Note:
        All parameters are calculated in pixels!!

    See also:
        :func:`skimage.measure.label`:
            Used to find different clouds. 
        :func:`skimage.measure.regionprops`:
            Used to calculate cloud properties.

    Parameters:
        cloudmask (ndarray): 2d binary cloud mask.
        connectivity (int):  Maximum number of orthogonal hops to consider
            a pixel/voxel as a neighbor (see :func:`skimage.measure.label`).

    Returns:
        list:
            List of :class:`RegionProperties`
            (see :func:`skimage.measure.regionprops`)
    """
    cloudmask[np.isnan(cloudmask)] = 0

    labels = measure.label(cloudmask, connectivity=connectivity)

    return measure.regionprops(labels)


def neighbor_distance(cloudmask, connectivity=1):
    """Calculate nearest neighbor distance for each cloud.

    Note: 
        Distance is given in pixels.

    See also: 
        :class:`scipy.spatial.cKDTree`:
            Used to calculate nearest neighbor distances. 

    Parameters: 
        cloudmask (ndarray): 2d binary cloud mask.
        connectivity (int):  Maximum number of orthogonal hops to consider
            a pixel/voxel as a neighbor (see :func:`skimage.measure.label`).

    Returns: 
        ndarray: Nearest neighbor distances in pixels.
    """
    cloudproperties = get_cloudproperties(cloudmask, connectivity=connectivity)
    
    centroids = [prop.centroid for prop in cloudproperties]
    indices = np.arange(len(centroids))
    neighbor_distance = np.zeros(len(centroids))
    centroids_array = np.asarray(centroids)

    for n, point in enumerate(centroids):
        # use all center of mass coordinates, but the one from the point
        mytree = sc.spatial.cKDTree(centroids_array[indices != n])
        dist, indexes = mytree.query(point)
        neighbor_distance[n] = dist

    return neighbor_distance


def iorg(cloudmask, connectivity=1):
    """Calculate the cloud cluster index 'I_org'.

    See also: 
        :func:`scipy.integrate.trapz`:
            Used to calculate the integral along the given axis using
            the composite trapezoidal rule.

    Parameters: 
        cloudmask (ndarray): 2d binary cloud mask.
        connectivity (int):  Maximum number of orthogonal hops to consider
            a pixel/voxel as a neighbor (see :func:`skimage.measure.label`).

    Returns:
        float: cloud cluster index I_org.

    References: 
        Tompkins, A. M., and A. G. Semie (2017), Organization of tropical 
        convection in low vertical wind shears: Role of updraft entrainment, 
        J. Adv. Model. Earth Syst., 9, 1046–1068, doi: 10.1002/2016MS000802.
        
    """
    nn = neighbor_distance(cloudmask, connectivity=connectivity)
    nn_sorted = np.sort(nn)
    
    nncdf = np.linspace(0, 1, len(nn))
    
    # theoretical nearest neighbor cumulative frequency
    # distribution (nncdf) of a random point process (Poisson)
    lamb = nn.size / cloudsmask.size
    nncdf_poisson = 1 - np.exp(-lamb * np.pi * nn_sorted**2)

    return sc.integrate.trapz(y=nncdf, x=nncdf_poisson)


def scai(cloudmask, connectivity=1):
    """Calculate the 'Simple Convective Aggregation Index (SCAI)'.  

    The SCAI is defined as the ratio of convective disaggregation
    to a potential maximal disaggregation.

    See also: 
        :func:`scipy.spatial.distance.pdist`:
            Used to calculate pairwise distances between cloud entities. 
        :func:`scipy.stats.mstats.gmean`:
            Used to calculate the geometric mean of all clouds in pairs. 

    Parameters:
        cloudmask (ndarray): 2d binary cloud mask.
        connectivity (int):  Maximum number of orthogonal hops to consider
            a pixel/voxel as a neighbor (see :func:`skimage.measure.label`).

    Returns:
        float: SCAI.

    References: 
        Tobin, I., S. Bony, and R. Roca, 2012: Observational Evidence for 
        Relationships between the Degree of Aggregation of Deep Convection, 
        Water Vapor, Surface Fluxes, and Radiation. J. Climate, 25, 6885–6904,
        https://doi.org/10.1175/JCLI-D-11-00258.1

    """
    cloudproperties = get_cloudproperties(cloudmask, connectivity=connectivity)
    centroids = [prop.centroid for prop in cloudproperties]

    # number of cloud clusters
    N = len(centroids)

    # potential maximum of N depending on cloud connectivity
    N_max = np.sum(~np.isnan(cloudmask)) / 2
    if connectivity == 2:
        N_max = np.sum(~np.isnan(cloudmask)) / 4

    # distance between points (center of mass of clouds) in pairs
    di = pdist(centroids, 'euclidean')
    # order-zero diameter
    D0 = sc.stats.mstats.gmean(di)

    # characteristic length of the domain (in pixels): diagonal of box
    L = np.sqrt(cloudmask.shape[0]**2 + cloudmask.shape[1]**2)

    return N / N_max * D0 / L * 1000


def cloudfraction(cloudmask):
    """Calculate cloud fraction based on cloud mask, while irnoring NaNs.
    
    Parameters:
        cloudmask (ndarray): 2d binary cloud mask.
        
    Returns:
        float: cloud fraction.
    """
    return np.nansum(cloudmask) / np.sum(~np.isnan(cloudmask))



from .common import NetCDF4, expects_file_info

__all__ = [
    'OceanRAIN',
]


class OceanRAIN(NetCDF4):
    """File handler that can read data from OceanRAIN NetCDF4 files.

    This object handles OceanRAIN V1.0 NetCDF4 files such as they are
    compatible with :mod:`typhon.collocations`, i.e.:

    * rename *latitude* and *longitude* field to *lat* and *lon*.
    * convert the content of the *time* variable to numpydtetime64 objects.

    Examples:

        Draw a world map with all measurements of OceanRAIN:

        .. :code-block:: python3

            from typhon.files import FileSet, OceanRAIN
            from typhon.plots import worldmap

            # Create a Dataset object that points to the files:
            ocean_rain = FileSet(
                "OceanRAIN_1.0/OceanRAIN__W__{ship}_{ship_id}__UHAM-ICDC__v1_0.nc",
                handler=OceanRAIN(),
                info_via="both",
            )

            # Create a figure:
            fig = plt.figure(figsize=(12, 12))
            ax = None

            # This iterates through all files from OceanRAIN and plots them:
            for file, data in ocean_rain.icollect(return_info=True):
                label = f"{file.attr['ship']} ({data['lat'].size} mins)"

                # We only plot every 1000th point to draw the worldmap faster:
                ax, plot = worldmap(
                    data["lat"][::1000], data["lon"][::1000], ax=ax,
                    s=6, label=label, background=True
                )

            # Add a nice legend
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(
                handles, labels, bbox_to_anchor=(0., 1.02, 1., .102), loc=3,
                ncol=2, borderaxespad=0., #mode="expand",
            )
            fig.show()

    """

    def __init__(self, **kwargs):
        """Initializes a OceanRAIN file handler class.

        Args:
            **kwargs: Additional key word arguments that are allowed for the
                :class:`typhon.spareice.handlers.common.NetCDF4` class.

        """
        # Call the base class initializer
        super().__init__(**kwargs)

    @expects_file_info()
    def get_info(self, file, **kwargs):
        """Get information about an OceanRAIN dataset file

        Args:
            file: A string containing path and name or a :class:`FileInfo`
                object of the file of which to get the information about.
            **kwargs: Additional keyword arguments.

        Returns:
            A :class:`FileInfo` object.
        """
        data = super().read(file, fields=("time",))

        file.times = [data["time"].min().item(0), data["time"].max().item(0)]

        return file

    @expects_file_info()
    def read(self, filename, **kwargs):
        """Read and parse a NetCDF file and load it to a xarray.Dataset

        Args:
            filename: Path and name of the file as string or FileInfo object.
            **kwargs: Additional key word arguments that are allowed for the
                :class:`~typhon.files.handlers.common.NetCDF4` class.

        Returns:
            A xarray.Dataset object.
        """

        # Make sure that the standard fields are always gonna be imported:
        fields = kwargs.pop("fields", None)
        if fields is not None:
            fields = {"time", "latitude", "longitude"} | set(fields)

        # This renaming makes the data compatible for collocate routines:
        mapping = {
            "latitude": "lat",
            "longitude": "lon",
            **kwargs.get("mapping", {})
        }

        return super().read(filename, fields=fields, mapping=mapping, **kwargs)



# -*- coding: utf-8 -*-
"""Handling of environment variables.

Implements a mapping object for environment variables.

The handler is used like ``os.environ``:

>>> typhon.environ['ARTS_DATA_PATH']
'path/to/data'
>>> typhon.environ.get('ARTS_DATA_PATH')
'path/to/data'

In addition to the user's environment, variables can be set in the in the
configuration file (:mod:`typhon.config`) in the ``environment`` section::

    [environment]
    ARTS_BUILD_PATH: /path/to/arts/build/

Note:
    If the environment variable is set explicitly, the value set in
    the configuration file is ignored.
"""
import os

from typhon.config import conf


__all__ = [
    'environ',
]


class _EnvironmentHandler:
    """A mapping object for environment variables.

    See module docstring of :mod:`typhon.environment` for more information.
    """
    def __getitem__(self, key):
        try:
            # First, try to return the value from the user's environment...
            return os.environ[key]
        except KeyError:
            # if the key is not set, try to find it in the config.
            # If this also fails, a KeyError is raised.
            return conf['environment'][key]

    def __setitem__(self, key, value):
        # If an environment variable is set, pass it to the actual user's
        # environment. This ensures consistent environments for subprocesses.
        os.environ[key] = value

    def __contains__(self, item):
        """Implement membership test operators.

        Returns:
            `True` if item is in self, `False` otherwise.
        """
        try:
            self[item]
            return True
        except KeyError:
            return False

    def get(self, key, default=None):
        """D.get(k[, d]) -> D[k] if k in D, else d.d defaults to None."""
        # Try to return the value from the user's environment.
        # If the key is not set, try to find it in the config.
        # If this also fails, return a default value.
        return os.environ.get(key, default=conf.get(section='environment',
                                                    option=key,
                                                    fallback=default)
                              )


environ = _EnvironmentHandler()



# -*- coding: utf-8 -*-

"""Utility functions related to plotting.
"""
import glob
import itertools
import os
import string

import matplotlib.pyplot as plt
import numpy as np
from typhon import constants


__all__ = [
    'center_colorbar',
    'figsize',
    'styles',
    'get_subplot_arrangement',
    'label_axes',
    'supcolorbar',
    'sorted_legend_handles_labels',
]


def center_colorbar(cb):
    """Center a diverging colorbar around zero.

    Convenience function to adjust the color limits of a colorbar. The function
    multiplies the absolute maximum of the data range by ``(-1, 1)`` and uses
    this range as new color limits.

    Note:
        The colormap used should be continuous. Resetting the clim for discrete
        colormaps may produce strange artefacts.

    Parameters:
        cb (matplotlib.colorbar.Colorbar): Colorbar to center.

    Examples:

    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        from typhon.plots import center_colorbar


        fig, ax = plt.subplots()
        sm = ax.pcolormesh(np.random.randn(10, 10) + 0.75, cmap='difference')
        cb = fig.colorbar(sm)
        center_colorbar(cb)

        plt.show()
    """
    # Set color limits to +- the absolute maximum of the data range.
    cb.set_clim(np.multiply((-1, 1), np.max(np.abs(cb.get_clim()))))


def figsize(w, portrait=False):
    """Return a figure size matching the golden ratio.

    This function takes a figure width and returns a tuple
    representing width and height in the golden ratio.
    Results can be returned for portrait orientation.

    Parameters:
        w (float): Figure width.
        portrait (bool): Return size for portrait format.

    Return:
        tuple: Figure width and size.

    Examples:
        >>> import typhon.plots
        >>> typhon.plots.figsize(1)
        (1, 0.61803398874989479)

        >>> typhon.plots.figsize(1, portrait=True)
        (1, 1.6180339887498949)
    """
    phi = constants.golden_ratio
    return (w, w * phi) if portrait else (w, w / phi)


def get_subplot_arrangement(n):
    """Get efficient (nrow, ncol) for n subplots

    If we want to put `n` subplots in a square-ish/rectangular
    arrangement, how should we arrange them?

    Returns (⌈√n⌉, ‖√n‖)
    """
    return (int(np.ceil(np.sqrt(n))),
            int(np.round(np.sqrt(n))))


class _StyleHandler:
    """Handle matplotlib stylesheets shipped with typhon.

    This handler serves several purposes:

        1. Return the absolute path to a typhon stylesheet:

        >>> from typhon.plots import styles
        >>> styles.get('typhon')
        '/PATH/typhon/typhon/plots/stylelib/typhon.mplstyle'

        or for backwards compatibility:

        >>> styles('typhon')
        '/PATH/typhon/typhon/plots/stylelib/typhon.mplstyle'

        2. Use style settings from a typhon stylesheet:

        >>> styles.use('typhon')

        You can also get/use several styles at a time:

        >>> styles.use(['typhon', 'typhon-dark'])

    Parameters:
        name (str): Style name.

    Returns:
        str: Absolute path to stylesheet.
    """
    def __init__(self):
        self.stylelib_dir = os.path.join(os.path.dirname(__file__), 'stylelib')

    def __call__(self, name=None):
        """Return absolute path to typhon stylesheet.

        Parameters:
            name (str, list): Style name or list of style names.
        """
        return self.get(name)

    def _get_style(self, name):
        """Return absolute path to a single typhon stylesheet."""
        style_path = os.path.join(self.stylelib_dir, name + '.mplstyle')

        if os.path.isfile(style_path):
            return style_path
        else:
            raise ValueError(
                f'"{name}" is not a valid style. '
                f'See ``styles.available`` for list of available styles.'
        )

    def get(self, name=None):
        """Return absolute path to typhon stylesheet.

        Parameters:
            name (str, list): Style name or list of style names.
        """
        if name is None:
            name = 'typhon'

        if isinstance(name, list):
            return [self._get_style(n) for n in name]
        elif isinstance(name, str):
            return self._get_style(name)
        else:
            raise TypeError('``name`` has to be of type str or list.')

    def use(self, name=None):
        """Use matplotlib style settings from a typhon style specification.

        Parameters:
            name (str, list): Style name or list of style names.
        """
        plt.style.use(self.get(name))

    @property
    def available(self):
        """Return list of available typhon styles."""
        pattern = os.path.join(self.stylelib_dir, '*.mplstyle')

        return [os.path.splitext(os.path.basename(s))[0]
                for s in glob.glob(pattern)]

styles = _StyleHandler()


def label_axes(axes=None, labels=None, loc=(.02, .9), **kwargs):
    """Walks through axes and labels each.

    Parameters:
        axes (iterable): An iterable container of :class:`AxesSubplot`.
        labels (iterable): Iterable of strings to use to label the axes.
            If ``None``, first upper and then lower case letters are used.
        loc (tuple of floats): Where to put the label in axes-fraction units.
        **kwargs: Additional keyword arguments are collected and
            passed to :func:`~matplotlib.pyplot.annotate`.

    Examples:
        .. plot::
            :include-source:

            import matplotlib.pyplot as plt
            from typhon.plots import label_axes, styles


            plt.style.use(styles('typhon'))

            # Automatic labeling of axes.
            fig, axes = plt.subplots(ncols=2, nrows=2)
            label_axes()

            # Manually specify the axes to label.
            fig, axes = plt.subplots(ncols=2, nrows=2)
            label_axes(axes[:, 0])  # label each row.

            # Pass explicit labels (and additional arguments).
            fig, axes = plt.subplots(ncols=2, nrows=2)
            label_axes(labels=map(str, range(axes.size)), weight='bold')

    .. Based on https://stackoverflow.com/a/22509497
    """

    # This code, or an earlier version, was posted by user 'tacaswell' on Stack
    # https://stackoverflow.com/a/22509497/974555 and is licensed under
    # CC-BY-SA 3.0.  This notice may not be removed.
    if axes is None:
        axes = plt.gcf().axes

    if labels is None:
        labels = string.ascii_uppercase + string.ascii_lowercase

    labels = itertools.cycle(labels)  # re-use labels rather than stop labeling

    for ax, lab in zip(axes, labels):
        ax.annotate(lab, xy=loc, xycoords='axes fraction', **kwargs)


def supcolorbar(mappable, fig=None, right=0.8, rect=(0.85, 0.15, 0.05, 0.7),
                **kwargs):
    """Create a common colorbar for all subplots in a figure.

    Parameters:
        mappable: A scalar mappable like to which the colorbar applies
            (e.g. :class:`~matplotlib.collections.QuadMesh`,
            :class:`~matplotlib.contour.ContourSet`, etc.).
        fig (:class:`~matplotlib.figure.Figure`):
            Figure to add the colorbar into.
        right (float): Fraction of figure to use for the colorbar (0-1).
        rect (array-like): Add an axes at postion
            ``rect = [left, bottom, width, height]`` where all quantities are
            in fraction of figure.
        **kwargs: Additional keyword arguments are passed to
            :func:`matplotlib.figure.Figure.colorbar`.

    Returns:
        matplotlib.colorbar.Colorbar: Colorbar.

    Note:
        In order to properly scale the value range of the colorbar the ``vmin``
        and ``vmax`` property should be set manually.

    Examples:
        .. plot::
            :include-source:

            import matplotlib.pyplot as plt
            import numpy as np
            from typhon.plots import supcolorbar


            fig, axes = plt.subplots(2, 2, sharex=True, sharey=True)
            for ax in axes.flat:
                sm = ax.pcolormesh(np.random.random((10,10)), vmin=0, vmax=1)

            supcolorbar(sm, label='Test label')

    """
    if fig is None:
        fig = plt.gcf()

    fig.subplots_adjust(right=right)
    cbar_ax = fig.add_axes(rect)

    return fig.colorbar(mappable, cax=cbar_ax, **kwargs)


def sorted_legend_handles_labels(ax=None, key=None, reverse=True):
    """Sort legend labels and handles.

    Returns legend handles and labels in descending order of y data peak
    values.

    Parameters:
        ax: Matplotlib axis.
        key (Callable): Function that takes :class:`~matplotlib.lines.Line2D`
                        object. See example below.
        reverse (bool): Default: True

    Returns:
        Tuple(handles, labels): Sorted legend handles and labels

    Example:
        >>> # Sort by maximum y value
        >>> ax.legend(*sorted_legend_handles_labels())

        >>> # Sort by minimum y value
        >>> ax.legend(*sorted_legend_handles_labels(
        ...           key=lambda line: np.min(line.get_ydata())))
    """
    if ax is None:
        ax = plt.gca()

    if key is None:
        def key(line):
            return np.max(line.get_ydata())

    # Sort legend entries by their cross section peak values
    return zip(*((h, l) for _, h, l in
                 sorted(
                     zip([key(h) for h in
                          ax.get_legend_handles_labels()[0]],
                         *ax.get_legend_handles_labels()),
                     reverse=reverse)))



# -*- coding: utf-8 -*-

"""This module contains functions to interact with ARTS.
"""

from typhon.arts import sensor  # noqa
from typhon.arts import xml  # noqa
from typhon.arts.common import *  # noqa

__all__ = [s for s in dir() if not s.startswith('_')]



import gc
import logging
import traceback
from collections import defaultdict
from datetime import datetime, timedelta
from multiprocessing import Process, Queue

import numpy as np
import pandas as pd
import xarray as xr

from typhon.geodesy import great_circle_distance
from typhon.geographical import GeoIndex
from typhon.utils import add_xarray_groups, get_xarray_groups
from typhon.utils.timeutils import to_datetime, to_timedelta, Timer

__all__ = [
    "Collocator",
    "check_collocation_data"
]

logger = logging.getLogger(__name__)


# The names for the processes. This started as an easter egg, but it actually
# helps to identify different processes during debugging.
PROCESS_NAMES = [
    'Newton', 'Einstein', 'Bohr', 'Darwin', 'Pasteur', 'Freud', 'Galilei',
    'Lavoisier', 'Kepler', 'Copernicus', 'Faraday', 'Maxwell', 'Bernard',
    'Boas', 'Heisenberg', 'Pauling', 'Virchow', 'Schrodinger', 'Rutherford',
    'Dirac', 'Vesalius', 'Brahe', 'Buffon', 'Boltzmann', 'Planck', 'Curie',
    'Herschel', 'Lyell', 'Laplace', 'Hubble', 'Thomson', 'Born', 'Crick',
    'Fermi', 'Euler', 'Liebig', 'Eddington', 'Harvey', 'Malpighi', 'Huygens',
    'Gauss', 'Haller', 'Kekule', 'Koch', 'Gell-Mann', 'Fischer', 'Mendeleev',
    'Glashow', 'Watson', 'Bardeen', 'Neumann', 'Feynman', 'Wegener', 'Hawking',
    'Leeuwenhoek', 'Laue', 'Kirchhoff', 'Bethe', 'Euclid', 'Mendel', 'Onnes',
    'Morgan', 'Helmholtz', 'Ehrlich', 'Mayr', 'Sherrington', 'Dobzhansky',
    'Delbruck', 'Lamarck', 'Bayliss', 'Chomsky', 'Sanger', 'Lucretius',
    'Dalton', 'Broglie', 'Linnaeus', 'Piaget', 'Simpson', 'Levi-Strauss',
    'Margulis', 'Landsteiner', 'Lorenz', 'Wilson', 'Hopkins', 'Elion', 'Selye',
    'Oppenheimer', 'Teller', 'Libby', 'Haeckel', 'Salk', 'Kraepelin',
    'Lysenko', 'Galton', 'Binet', 'Kinsey', 'Fleming', 'Skinner', 'Wundt',
    'Archimedes'
]


class ProcessCrashed(Exception):
    """Helper exception for crashed processes"""
    pass


class Collocator:
    def __init__(
            self, threads=None, name=None, #log_dir=None
    ):
        """Initialize a collocator object that can find collocations

        Args:
            threads: Finding collocations can be parallelized in threads. Give
                here the maximum number of threads that you want to use. Which
                number of threads is the best, may be machine-dependent. So
                this is a parameter that you can use to fine-tune the
                performance. Note: Not yet implemented due to GIL usage of
                sklearn BallTree.
            name: The name of this collocator, will be used in log statements.
        """

        self.empty = None  # xr.Dataset()

        self.index = None
        self.index_with_primary = False

        self.threads = threads

        # These optimization parameters will be overwritten in collocate
        self.bin_factor = None
        self.magnitude_factor = None
        self.tunnel_limit = None
        self.leaf_size = None

        self.name = name if name is not None else "Collocator"

    # If no collocations are found, this will be returned. We need empty
    # arrays to concatenate the results without problems:
    @property
    def no_pairs(self):
        return np.array([[], []])

    @property
    def no_intervals(self):
        return np.array([], dtype='timedelta64[ns]')

    @property
    def no_distances(self):
        return np.array([])

    def __call__(self, *args, **kwargs):
        return self.collocate(*args, **kwargs)

    def _debug(self, msg):
        logger.debug(f"[{self.name}] {msg}")

    def _info(self, msg):
        logger.info(f"[{self.name}] {msg}")

    def _error(self, msg):
        logger.error(f"[{self.name}] {msg}")

    def collocate_filesets(
            self, filesets, start=None, end=None, processes=None, output=None,
            bundle=None, skip_file_errors=False, post_processor=None,
            post_processor_kwargs=None, **kwargs
    ):
        """Find collocation between the data of two filesets

        If you want to save the collocations directly to disk, it may be easier
        to use :meth:`~typhon.collocations.Collocations.search` directly.

        Args:
            filesets: A list of two :class:`FileSet` objects, the primary and
                the secondary fileset. Can be also
                :class:`~typhon.collocations.common.Collocations` objects with
                `read_mode=collapse`. The order of the filesets is irrelevant
                for the results of the collocation search but files from the
                secondary fileset might be read multiple times if  using
                parallel processing (`processes` is greater than one). The
                number of output files could be different (see also the option
                `bundle`).
            start: Start date either as datetime object or as string
                ("YYYY-MM-DD hh:mm:ss"). Year, month and day are required.
                Hours, minutes and seconds are optional. If not given, it is
                datetime.min per default.
            end: End date. Same format as "start". If not given, it is
                datetime.max per default.
            processes: Collocating can be parallelized which improves the
                performance significantly. Pass here the number of processes to
                use.
            output: Fileset object where the collocated data should be stored.
            bundle: Set this to *primary* if you want to bundle the output 
                files by their collocated primaries, i.e. there will be only 
                one output file per primary. *daily* is also possible, then all
                files from one day are bundled together. Per default, all
                collocations for each file match will be saved separately.
                This might lead to a high number of output files.
                Note: *daily* means one process bundles all collocations from
                one day into one output file. If using multiple processes, this
                could still produce several daily output files per day.
            skip_file_errors: If this is *True* and a file could not be read, 
                the file and its match will be skipped and a warning will be 
                printed. Otheriwse the program will stop (default).
            post_processor: A function for post-processing the collocated data
                before saving it to `output`. Must accept two parameters: a
                xarray.Dataset with the collocated data and a dictionary with
                the path attributes from the collocated files.
            post_processor_kwargs: A dictionary with keyword arguments that
                should be passed to `post_processor`.
            **kwargs: Further keyword arguments that are allowed for
                :meth:`collocate`.

        Yields:
            A xarray.Dataset with the collocated data if `output` is not set.
            If `output` is set to a FileSet-like object, only the filename of
            the stored collocations is yielded. The results are not ordered if
            you use more than one process. For more information about the 
            yielded xarray.Dataset have a look at :meth:`collocate`.

        Examples:

        .. code-block:: python

        """
        timer = Timer().start()

        if len(filesets) != 2:
            raise ValueError("Only collocating two filesets at once is allowed"
                             "at the moment!")

        # Check the max_interval argument because we need it later
        max_interval = kwargs.get("max_interval", None)
        if max_interval is None:
            raise ValueError("Collocating filesets without max_interval is"
                             " not yet implemented!")

        if start is None:
            start = datetime.min
        else:
            start = to_datetime(start)
        if end is None:
            end = datetime.max
        else:
            end = to_datetime(end)

        self._info(f"Collocate from {start} to {end}")

        # Find the files from both filesets which overlap tempoerally.
        matches = list(filesets[0].match(
            filesets[1], start=start, end=end, max_interval=max_interval,
        ))

        if processes is None:
            processes = 1

        # Make sure that there are never more processes than matches
        processes = min(processes, len(matches))

        total_matches = sum(len(match[1]) for match in matches)
        self._info(f"using {processes} process(es) on {total_matches} matches")

        # MAGIC with processes
        # Each process gets a list with matches. Important: the matches should
        # be continuous to guarantee a good performance. After finishing one
        # match, the process pushes its results to the result queue. If errors
        # are raised during collocating, the raised errors are pushed to the 
        # error queue,
        matches_chunks = np.array_split(matches, processes)

        # This queue collects all results:
        results = Queue(maxsize=processes)

        # This queue collects all error exceptions
        errors = Queue()

        # Extend the keyword arguments that we are going to pass to
        # _collocate_files:
        kwargs.update({
            "start": start,
            "end": end,
            "filesets": filesets,
            "output": output,
            "bundle": bundle,
            "skip_file_errors": skip_file_errors,
            "post_processor": post_processor,
            "post_processor_kwargs": post_processor_kwargs,
        })

        # This list contains all running processes
        process_list = [
            Process(
                target=Collocator._process_caller,
                args=(
                    self, results, errors, PROCESS_NAMES[i],
                ),
                kwargs={**kwargs, "matches": matches_chunk},
                daemon=True,
            )
            for i, matches_chunk in enumerate(matches_chunks)
        ]

        # We want to keep track of the progress of the collocation search since
        # it may take a while.
        process_progress = {
            name: 0.  # Each process is at 0 percent at the beginning
            for name in PROCESS_NAMES[:processes]
        }

        # Start all processes:
        for process in process_list:
            process.start()

        # As long as some processes are still running, wait for their results:
        running = process_list.copy()

        processed_matches = 0

        # The main process has two tasks during its child processes are
        # collocating. 
        # 1) Collect their results and yield them to the user
        # 2) Display the progress and estimate the remaining processing time
        while running:

            # Filter out all processes that are dead: they either crashed or
            # complete their task
            running = [
                process for process in running if process.is_alive()
            ]

            # Get all results from the result queue
            while not results.empty():
                process, progress, result = results.get()

                # The process might be crashed. To keep the remaining time
                # estimation useful, we exclude the crashed process from the
                # calculation.
                if result is ProcessCrashed:
                    del process_progress[process]
                else:
                    process_progress[process] = progress

                try:
                    nerrors = errors.qsize()
                except NotImplementedError:
                    nerrors = 'unknown'

                self._print_progress(
                    timer.elapsed, process_progress, len(running), nerrors)
                if result is not None:
                    yield result

                # Explicit free up memory:
                gc.collect()

        for process in process_list:
            process.join()

        if not errors.empty():
            self._error("Some processes terminated due to errors:")

        while not errors.empty():
            error = errors.get()
            msg = '\n'.join([
                "-"*79,
                error[2],
                "".join(traceback.format_tb(error[1])),
                "-" * 79 + "\n"
            ])
            self._error(msg)

    @staticmethod
    def _print_progress(elapsed_time, process_progress, processes, errors):

        elapsed_time -= timedelta(microseconds=elapsed_time.microseconds)
        if len(process_progress) == 0:
            msg = "-"*79 + "\n"
            msg += f"100% | {elapsed_time} hours elapsed | " \
                   f"{errors} processes failed\n"
            msg += "-"*79 + "\n"
            logger.error(msg)
            return

        progress = sum(process_progress.values()) / len(process_progress)

        try:
            expected_time = elapsed_time * (100 / progress - 1)
            expected_time -= timedelta(
                    microseconds=expected_time.microseconds)
        except ZeroDivisionError:
            expected_time = "unknown"

        msg = "-"*79 + "\n"
        msg += f"{progress:.0f}% | {elapsed_time} hours elapsed, " \
               f"{expected_time} hours left | {processes} proc running, " \
               f"{errors} failed\n"
        msg += "-"*79 + "\n"
        logger.error(msg)

    @staticmethod
    def _process_caller(
            self, results, errors, name, output, bundle, post_processor,
            post_processor_kwargs, **kwargs):
        """Wrapper around _collocate_matches

        This function is called for each process. It communicates with the main
        process via the result and error queue.

        Result Queue:
            Adds for each collocated file match the process name, its progress
            and the actual results.

        Error Queue:
            If an error is raised, the name of this proces and the error
            messages is put to this queue.
        """
        self.name = name

        # We keep track of how many file pairs we have already processed to
        # make the error debugging easier. We need the match in flat form:
        matches = [
            [match[0], secondary]
            for match in kwargs['matches']
            for secondary in match[1]
        ]

        # If we want to bundle the output, we need to collect some contents.
        # The current_bundle_tag stores a certain information for the current
        # bundle (e.g. filename of primary or day of the year). If it changes,
        # the bundle is stored to disk and a new bundle is created.
        cached_data = []
        cached_attributes = {}
        current_bundle_tag = None
        try:
            processed = 0
            collocated_matches = self._collocate_matches(**kwargs)
            for collocations, attributes in collocated_matches:
                match = matches[processed]
                processed += 1
                progress = 100 * processed / len(matches)

                if collocations is None:
                    results.put([name, progress, None])
                    continue

                # The user does not want to bundle anything therefore just save
                # the current collocations
                if bundle is None:
                    result = self._save_and_return(
                            collocations, attributes, output,
                            post_processor, post_processor_kwargs
                    )
                    results.put([name, progress, result])
                    continue

                # The user may want to bundle the collocations before writing
                # them to disk, e.g. by their primaries.
                save_cache = self._should_save_cache(
                        bundle, current_bundle_tag, match,
                        to_datetime(collocations.attrs["start_time"])
                )

                if save_cache:
                    result = self._save_and_return(
                        cached_data,
                        cached_attributes, output,
                        post_processor, post_processor_kwargs
                    )
                    results.put([name, progress, result])

                    cached_data = []
                    cached_attributes = {}

                # So far, we have not cached any collocations or we still need
                # to wait before saving them to disk.
                cached_data.append(collocations)
                cached_attributes.update(**attributes)

                if bundle == "primary":
                    current_bundle_tag = match[0].path
                elif bundle == "daily":
                    current_bundle_tag = \
                        to_datetime(collocations.attrs["start_time"]).date()

            # After all iterations, save last cached data to disk:
            if cached_data:
                result = self._save_and_return(
                    cached_data,
                    cached_attributes, output,
                    post_processor, post_processor_kwargs
                )
                results.put([name, progress, result])

        except Exception as exception:
            # Tell the main process to stop considering this process for the 
            # remaining processing time:
            results.put(
                [name, 100., ProcessCrashed]
            )

            self._error("ERROR: I got a problem and terminate!")

            # Build a message that contains all important information for
            # debugging:

            msg = f"Process {name} ({matches[0][0].times[0]} -" \
                f"{matches[-1][0].times[1]}) failed\n" \
                f"Failed to collocate {matches[processed]} with"\
                f"{matches[processed]}\n"

            # The main process needs to know about this exception!
            error = [
                name, exception.__traceback__,
                msg + "ERROR: " + str(exception)
            ]
            errors.put(error)

            self._error(exception)

            # Finally, raise the exception to terminate this process:
            raise exception

        self._info(f"Finished all {len(matches)} matches")

    def _save_and_return(self, collocations, attributes, output,
            post_processor, post_processor_kwargs):
        """Save collocations to disk or return them"""

        if isinstance(collocations, list):
            collocations = concat_collocations(
                    collocations
            )

        if output is None:
            return collocations, attributes
        else:
            filename = output.get_filename(
                [to_datetime(collocations.attrs["start_time"]),
                 to_datetime(collocations.attrs["end_time"])],
                fill=attributes
            )

            # Apply a post processor function from the user
            if post_processor is not None:
                if post_processor_kwargs is None:
                    post_processor_kwargs = {}

                collocations = post_processor(
                    collocations, attributes, **post_processor_kwargs
                )

            if collocations is None:
                return None

            self._info(f"Store collocations to\n{filename}")

            # Write the data to the file.
            output.write(collocations, filename)
            return filename

    @staticmethod
    def _should_save_cache(bundle, current_bundle_tag, match, start_time):
        """Return true if the cache should be saved otherwise false
        """
        if current_bundle_tag is None:
            return False
        elif bundle == "primary":
            # Check whether the primary has changed since the last time:
            return current_bundle_tag != match[0].path
        elif bundle == "daily":
            # Has the day changed since last time?
            return current_bundle_tag != start_time.date()

        # In all other cases, the bundle should not be saved yet:
        return False

    def _collocate_matches(
        self, filesets, matches, skip_file_errors, **kwargs
    ):
        """Load file matches and collocate their content

        Yields:
            A tuple of two items: the first is always the current percentage
            of progress. If output is True, the second is only the filename of
            the saved collocations. Otherwise, it is a tuple of collocations
            and their collected :class:`~typhon.files.handlers.common.FileInfo`
            attributes as a dictionary.
        """
        # Load all matches in a parallized queue:
        loaded_matches = filesets[0].align(
            filesets[1], matches=matches, return_info=True, compact=False,
            skip_errors=skip_file_errors,
        )


        for loaded_match in loaded_matches:
            # The FileInfo objects of the matched files:
            files = loaded_match[0][0], loaded_match[1][0]

            # We copy the data from the matches since they might be used for
            # other matches as well:
            primary, secondary = \
                loaded_match[0][1].copy(), loaded_match[1][1].copy()

            self._debug(f"Collocate {files[0].path}\nwith {files[1].path}")

            collocations = self.collocate(
                (filesets[0].name, primary),
                (filesets[1].name, secondary), **kwargs,
            )

            if collocations is None:
                self._debug("Found no collocations!")
                # At least, give the process caller a progress update:
                yield None, None
                continue

            # Check whether the collocation data is compatible and was build
            # correctly
            check_collocation_data(collocations)

            found = [
                collocations[f"{filesets[0].name}/time"].size,
                collocations[f"{filesets[1].name}/time"].size
            ]

            self._debug(
                f"Found {found[0]} ({filesets[0].name}) and "
                f"{found[1]} ({filesets[1].name}) collocations"
            )

            # Add the names of the processed files:
            for f in range(2):
                if f"{filesets[f].name}/__file" in collocations.variables:
                    continue

                collocations[f"{filesets[f].name}/__file"] = files[f].path

            # Collect the attributes of the input files. The attributes get a
            # prefix, primary or secondary, to allow not-unique names.
            attributes = {
                f"primary.{p}" if f == 0 else f"secondary.{p}": v
                for f, file in enumerate(files)
                for p, v in file.attr.items()
            }

            yield collocations, attributes


    def collocate(
            self, primary, secondary, max_interval=None, max_distance=None,
            bin_factor=1, magnitude_factor=10, tunnel_limit=None, start=None,
            end=None, leaf_size=40
    ):
        """Find collocations between two xarray.Dataset objects

        Collocations are two or more data points that are located close to each
        other in space and/or time.

        Each xarray.Dataset contain the variables *time*, *lat*, *lon*. They
        must be - if they are coordinates - unique. Otherwise, their
        coordinates must be unique, i.e. they cannot contain duplicated values.
        *time* must be a 1-dimensional array with a *numpy.datetime64*-like
        data type. *lat* and *lon* can be gridded, i.e. they can be multi-
        dimensional. However, they must always share the first dimension with
        *time*. *lat* must be latitudes between *-90* (south) and *90* (north)
        and *lon* must be longitudes between *-180* (west) and *180* (east)
        degrees. See below for examples.

        The collocation searched is performed with a fast ball tree
        implementation by scikit-learn. The ball tree is cached and reused
        whenever the data points from `primary` or `secondary` have not
        changed.

        If you want to find collocations between FileSet objects, use
        :class:`collocate_filesets` instead.

        Args:
            primary: A tuple of a string with the dataset name and a
                xarray.Dataset that fulfill the specifications from above. Can
                be also a xarray.Dataset only, the name is then automatically
                set to *primary*.
            secondary: A tuple of a string with the dataset name and a
                xarray.Dataset that fulfill the specifications from above. Can
                be also a xarray.Dataset only, the name is then automatically
                set to *secondary*.
            max_interval: Either a number as a time interval in seconds, a
                string containing a time with a unit (e.g. *100 minutes*) or a
                timedelta object. This is the maximum time interval between two
                data points. If this is None, the data will be searched for
                spatial collocations only.
            max_distance: Either a number as a length in kilometers or a string
                containing a length with a unit (e.g. *100 meters*). This is
                the maximum distance between two data points to meet the
                collocation criteria. If this is None, the data will be
                searched for temporal collocations only. Either `max_interval`
                or *max_distance* must be given.
            tunnel_limit: Maximum distance in kilometers at which to switch
                from tunnel to haversine distance metric. Per default this
                algorithm uses the tunnel metric, which simply transform all
                latitudes and longitudes to 3D-cartesian space and calculate
                their euclidean distance. This is faster than the haversine
                metric but produces an error that grows with larger distances.
                When searching for distances exceeding this limit
                (`max_distance` is greater than this parameter), the haversine
                metric is used, which is more accurate but takes more time.
                Default is 1000 kilometers.
            magnitude_factor: Since building new trees is expensive, this
                algorithm tries to use the last tree when possible (e.g. for
                data with fixed grid). However, building the tree with the
                larger dataset and query it with the smaller dataset is faster
                than vice versa. Depending on which premise to follow, there
                might have a different performance in the end. This parameter
                is the factor of that one dataset must be larger than the other
                to throw away an already-built ball tree and rebuild it with
                the larger dataset.
            leaf_size: The size of one leaf in the Ball Tree. The higher the
                leaf size the faster is the tree building but the slower is the
                tree query. The optimal leaf size is dataset-dependent. Default
                is 40.
            bin_factor: When using a temporal criterion via `max_interval`, the
                data will be temporally binned to speed-up the search. The bin
                size is `bin_factor` * `max_interval`. Which bin factor is the
                best, may be dataset-dependent. So this is a parameter that you
                can use to fine-tune the performance.
            start: Limit the collocated data from this start date. Can be
                either as datetime object or as string ("YYYY-MM-DD hh:mm:ss").
                Year, month and day are required. Hours, minutes and seconds
                are optional. If not given, it is datetime.min per default.
            end: End date. Same format as "start". If not given, it is
                datetime.max per default.

        Returns:
            None if no collocations were found. Otherwise,
            a xarray.Dataset with the collocated data in *compact* form. It
            consists of three groups (groups of variables containing */* in
            their name): the *primary*, *secondary* and the *Collocations*
            group. If you passed `primary` or `secondary` with own names,
            they will be used in the output. The *Collocations* group contains
            information about the found collocations. *Collocations/pairs* is
            a 2xN array where N is the number of found collocations. It
            contains the indices of the *primary* and *secondary* data points
            which are collocations. The indices refer to the data points stored
            in the *primary* or *secondary* group. *Collocations/interval* and
            *Collocations/distance* are the intervals and distances between the
            collocations in seconds and kilometers, respectively. Collocations
            in *compact* form are efficient when saving them to disk but it
            might be complicated to use them directly. Consider applying
            :func:`~typhon.collocations.common.collapse` or
            :func:`~typhon.collocations.common.expand` on them.

        Examples:

            .. code-block: python
                
                # TODO: Update this example!

                import numpy as np
                from typhon.collocations import Collocator

                # Create the data. primary and secondary can also be
                # xarray.Dataset objects:
                primary = {
                    "time": np.arange(
                        "2018-01-01", "2018-01-02", dtype="datetime64[h]"
                    ),
                    "lat": 30.*np.sin(np.linspace(-3.14, 3.14, 24))+20,
                    "lon": np.linspace(0, 90, 24),
                }
                secondary = {
                    "time": np.arange(
                        "2018-01-01", "2018-01-02", dtype="datetime64[h]"
                    ),
                    "lat": 30.*np.sin(np.linspace(-3.14, 3.14, 24)+1.)+20,
                    "lon": np.linspace(0, 90, 24),
                }

                # Find collocations with a maximum distance of 300 kilometers
                # and a maximum interval of 1 hour
                collocator = Collocator()
                collocated = collocator.collocate(
                    primary, secondary,
                    max_distance="300km", max_interval="1h"
                )

                print(collocated)


        """
        if max_distance is None and max_interval is None:
            raise ValueError(
                "Either max_distance or max_interval must be given!"
            )

        if max_interval is not None:
            max_interval = to_timedelta(max_interval, numbers_as="seconds")

        # The user can give strings instead of datetime objects:
        start = datetime.min if start is None else to_datetime(start)
        end = datetime.max if end is None else to_datetime(end)

        # Did the user give the datasets specific names?
        primary_name, primary, secondary_name, secondary = self._get_names(
            primary, secondary
        )

        # Select the common time period of both datasets and flat them.
        primary, secondary = self._prepare_data(
            primary, secondary, max_interval, start, end
        )

        # Maybe there is no data left after selection?
        if primary is None:
            return self.empty

        self.bin_factor = bin_factor
        self.magnitude_factor = magnitude_factor
        self.tunnel_limit = tunnel_limit
        self.leaf_size = leaf_size

        timer = Timer().start()
        
        # We cannot allow NaNs in the time, lat or lon fields
        not_nans1 = self._get_not_nans(primary)
        not_nans2 = self._get_not_nans(secondary)

        # Retrieve the important fields from the data. To avoid any overhead by
        # xarray, we use the plain numpy.arrays and do not use the isel method
        # (see https://github.com/pydata/xarray/issues/2227). We rather use
        # index arrays that we use later to select the rest of the data
        lat1 = primary.lat.values[not_nans1]
        lon1 = primary.lon.values[not_nans1]
        time1 = primary.time.values[not_nans1]
        lat2 = secondary.lat.values[not_nans2]
        lon2 = secondary.lon.values[not_nans2]
        time2 = secondary.time.values[not_nans2]
        original_indices = [
            np.arange(primary.time.size)[not_nans1],
            np.arange(secondary.time.size)[not_nans2]
        ]
        self._debug(f"{timer} for filtering NaNs")

        # We can search for spatial collocations (max_interval=None), temporal
        # collocations (max_distance=None) or both.
        if max_interval is None:
            # Search for spatial collocations only:
            pairs, distances = self.spatial_search(
                lat1, lon1, lat2, lon2, max_distance,
            )

            intervals = self._get_intervals(
                time1[pairs[0]], time2[pairs[1]]
            )

            return self._create_return(
                primary, secondary, primary_name, secondary_name,
                self._to_original(pairs, original_indices),
                intervals, distances,
                max_interval, max_distance
            )
        elif max_distance is None:
            # Search for temporal collocations only:
            pairs, intervals = self.temporal_search(
                time1, time2, max_interval
            )

            distances = self._get_distances(
                lat1[pairs[0]], lon1[pairs[0]],
                lat2[pairs[1]], lon2[pairs[1]],
            )

            return self._create_return(
                primary, secondary, primary_name, secondary_name,
                self._to_original(pairs, original_indices),
                intervals, distances,
                max_interval, max_distance
            )

        # The user wants to use both criteria and search for spatial and
        # temporal collocations. At first, we do a coarse temporal pre-binning
        # so that we only search for collocations between points that might
        # also be temporally collocated. Unfortunately, this also produces an
        # overhead that is only negligible if we have a lot of data:
        data_magnitude = time1.size * time2.size

        if data_magnitude > 100_0000:
            # We have enough data, do temporal pre-binning!
            pairs, distances = self.spatial_search_with_temporal_binning(
                {"lat": lat1, "lon": lon1, "time": time1},
                {"lat": lat2, "lon": lon2, "time": time2},
                max_distance, max_interval
            )
        else:
            # We do not have enough data to justify that whole pre-binning.
            # Simply do it directly!
            pairs, distances = self.spatial_search(
                lat1, lon1, lat2, lon2, max_distance,
            )

        # Did we find any spatial collocations?
        if not pairs.any():
            return self.empty

        # Check now whether the spatial collocations really pass the temporal
        # condition:
        passed_temporal_check, intervals = self._temporal_check(
            time1[pairs[0]], time2[pairs[1]], max_interval
        )

        # Return only the values that passed the time check
        return self._create_return(
            primary, secondary, primary_name, secondary_name,
            self._to_original(
                pairs[:, passed_temporal_check], original_indices),
            intervals, distances[passed_temporal_check],
            max_interval, max_distance
        )

    @staticmethod
    def _to_original(pairs, original_indices):
        return np.array([
            original_indices[i][pair_array]
            for i, pair_array in enumerate(pairs)
        ])

    @staticmethod
    def _get_names(primary, secondary):
        # Check out whether the user gave the primary and secondary any name:
        if isinstance(primary, (tuple, list)):
            primary_name, primary = primary
        else:
            primary_name = "primary"
        if isinstance(secondary, (tuple, list)):
            secondary_name, secondary = secondary
        else:
            secondary_name = "secondary"

        return primary_name, primary, secondary_name, secondary

    def _prepare_data(self, primary, secondary, max_interval, start, end):
        """Prepare the data for the collocation search
        
        This method selects the time period which should be searched for
        collocations and flats the input datasets if they have gridded
        variables.

        Returns:
            The datasets constraint to the common time period, sorted by time
            and flattened. If no common time period could be found, two None
            objects are returned.
        """
        if max_interval is not None:
            timer = Timer().start()
            # We do not have to collocate everything, just the common time
            # period expanded by max_interval and limited by the global start
            # and end parameter:
            primary_period, secondary_period = self._get_common_time_period(
                primary, secondary, max_interval, start, end
            )

            # Check whether something is left:
            if not primary_period.size or not secondary_period.size:
                return None, None

            # We need everything sorted by the time, otherwise xarray's stack
            # method makes problems:
            primary_period = primary_period.sortby(primary_period)
            primary_dim = primary_period.dims[0]
            secondary_period = secondary_period.sortby(secondary_period)
            secondary_dim = secondary_period.dims[0]

            # Select the common time period and while using sorted indices:
            primary = primary.sel(**{primary_dim: primary_period[primary_dim]})
            secondary = secondary.sel(
                **{secondary_dim: secondary_period[secondary_dim]}
            )

            # Check whether something is left:
            if not primary_period.size or not secondary_period.size:
                return None, None

            self._debug(f"{timer} for selecting common time period")

        # Flat the data: For collocating, we need a flat data structure.
        # Fortunately, xarray provides the very convenient stack method
        # where we can flat multiple dimensions to one. Which dimensions do
        # we have to stack together? We need the fields *time*, *lat* and
        # *lon* to be flat. So we choose their dimensions to be stacked.
        timer = Timer().start()
        primary = self._flat_to_main_coord(primary)
        secondary = self._flat_to_main_coord(secondary)
        self._debug(f"{timer} for flatting data")

        return primary, secondary

    @staticmethod
    def _get_common_time_period(
            primary, secondary, max_interval, start, end):
        max_interval = pd.Timedelta(max_interval)

        # We want to select a common time window from both datasets,
        # aligned to the primary's time coverage. Because xarray has a
        # very annoying bug in time retrieving
        # (https://github.com/pydata/xarray/issues/1240), this is a
        # little bit cumbersome:
        common_start = max(
            start,
            pd.Timestamp(primary.time.min().item(0)) - max_interval,
            pd.Timestamp(secondary.time.min().item(0)) - max_interval
        )
        common_end = min(
            end,
            pd.Timestamp(primary.time.max().item(0)) + max_interval,
            pd.Timestamp(secondary.time.max().item(0)) + max_interval
        )

        primary_period = primary.time.where(
            (primary.time.values >= np.datetime64(common_start))
            & (primary.time.values <= np.datetime64(common_end))
        ).dropna(primary.time.dims[0])

        secondary_period = secondary.time.where(
            (secondary.time.values >= np.datetime64(common_start))
            & (secondary.time.values <= np.datetime64(common_end))
        ).dropna(secondary.time.dims[0])

        return primary_period, secondary_period

    @staticmethod
    def _get_not_nans(dataset):
        return dataset.lat.notnull().values & dataset.lon.notnull().values

    @staticmethod
    def _flat_to_main_coord(data):
        """Make the dataset flat despite of its original structure

        We need a flat dataset structure for the collocation algorithms, i.e.
        time, lat and lon are not allowed to be gridded, they must be
        1-dimensional and share the same dimension (namely *collocation*).
        There are three groups of original data structures that this method
        can handle:

        * linear (e.g. ship track measurements): time, lat and lon have the
            same dimension and are all 1-dimensional. Fulfills all criteria
            from above. No action has to be taken.
        * gridded_coords (e.g. instruments on satellites with gridded swaths):
            lat or lon are gridded (they have multiple dimensions). Stack
            the coordinates of them together to a new shared dimension.

        Args:
            data: xr.Dataset object

        Returns:
            A xr.Dataset where time, lat and lon are aligned on one shared
            dimension.
        """
        # Flat:
        shared_dims = list(
            set(data.time.dims) | set(data.lat.dims) | set(data.lon.dims)
        )

        # Check whether the dataset is flat (time, lat and lon share the same
        # dimension size and are 1-dimensional)
        if len(shared_dims) == 1:

            if shared_dims[0] in ("time", "lat", "lon"):
                # One of the key variables is the main dimension! Change this:
                data["collocation"] = shared_dims[0], np.arange(
                    data[shared_dims[0]].size)
                data = data.swap_dims({shared_dims[0]: "collocation"})
                data = data.reset_coords(shared_dims[0])

                # So far, collocation is a coordinate. We want to make it to a
                # dimension,  so drop its values:
                return data.drop("collocation")

            return data.rename({
                shared_dims[0]: "collocation"
            })

        # The coordinates are gridded:
        # Some field might be more deeply stacked than another. Choose the
        # dimensions of the most deeply stacked variable:
        dims = max(
            data["time"].dims, data["lat"].dims, data["lon"].dims,
            key=lambda x: len(x)
        )
        # We want to be able to retrieve additional fields after collocating.
        # Therefore, we give each dimension that is no coordinate yet a value
        # to use them as indices later.
        for dim in dims:
            if dim not in data.coords:
                data[dim] = dim, np.arange(data.dims[dim])

        # We assume that coordinates must be unique! Otherwise, we would have
        # to use this ugly work-around:
        # Replace the former coordinates with new coordinates that have unique 
        # values.
        # new_dims = []
        # for dim in dims:
        #     new_dim = f"__replacement_{dim}"
        #     data[new_dim] = dim, np.arange(data.dims[dim])
        #     data = data.swap_dims({dim: new_dim})
        #     new_dims.append(new_dim)
        return data.stack(collocation=dims)

    def _create_return(
            self, primary, secondary, primary_name, secondary_name,
            original_pairs, intervals, distances,
            max_interval, max_distance
    ):
        if not original_pairs.any():
            return self.empty

        pairs = []
        output = {}

        names = [primary_name, secondary_name]
        for i, dataset in enumerate([primary, secondary]):
            # name of the current dataset (primary or secondary)
            name = names[i]

            # These are the indices of the points in the original data that
            # have collocations. We remove the duplicates since we want to copy
            # the required data only once. They are called original_indices
            # because they are the indices in the original data array:
            original_indices = pd.unique(original_pairs[i])

            # After selecting the collocated data, the original indices cannot
            # be applied any longer. We need new indices that indicate the
            # pairs in the collocated data.
            new_indices = np.empty(original_indices.max() + 1, dtype=int)
            new_indices[original_indices] = np.arange(
                original_indices.size
            )

            collocation_indices = new_indices[original_pairs[i]]

            # Save the collocation indices in the metadata group:
            pairs.append(collocation_indices)

            output[names[i]] = dataset.isel(collocation=original_indices)

            # We have to convert the MultiIndex to a normal index because we
            # cannot store it to a file otherwise. We can convert it by simply
            # setting it to new values, but we are losing the sub-level
            # coordinates (the dimenisons that we stacked to create the
            # multi-index in the first place) with that step. Hence, we store
            # the sub-level coordinates in additional dataset to preserve them.
            main_coord_is_multiindex = isinstance(
                output[name].get_index("collocation"),
                pd.core.indexes.multi.MultiIndex
            )
            if main_coord_is_multiindex:
                stacked_dims_data = xr.merge([
                    xr.DataArray(
                        output[name][dim].values,
                        name=dim, dims=["collocation"]
                    )
                    for dim in output[name].get_index("collocation").names
                ])

            # Okay, actually we want to get rid of the main coordinate. It
            # should stay as a dimension name but without own labels. I.e. we
            # want to drop it. Because it still may a MultiIndex, we cannot
            # drop it directly but we have to set it to something different.
            output[name]["collocation"] = \
                np.arange(output[name]["collocation"].size)

            if main_coord_is_multiindex:
                # Now, since we unstacked the multi-index, we can add the
                # stacked dimensions back to the dataset:
                output[name] = xr.merge(
                    [output[name], stacked_dims_data],
                )

            # For the flattening we might have created temporal variables,
            # also collect them to drop:
            vars_to_drop = [
                var for var in output[name].variables.keys()
                if var.startswith("__replacement_")
            ]

            output[name] = output[name].drop([
                f"collocation", *vars_to_drop
            ])

        # Merge all datasets into one:
        output = add_xarray_groups(
            xr.Dataset(), **output
        )

        # This holds the collocation information (pairs, intervals and
        # distances):
        metadata = xr.Dataset()
        metadata["pairs"] = xr.DataArray(
            np.array(pairs, dtype=int), dims=("group", "collocation"),
            attrs={
                "max_interval": f"Max. interval in secs: {max_interval}",
                "max_distance": f"Max. distance in kilometers: {max_distance}",
                "primary": primary_name,
                "secondary": secondary_name,
            }
        )
        metadata["interval"] = xr.DataArray(
            intervals, dims=("collocation", ),
            attrs={
                "max_interval": f"Max. interval in secs: {max_interval}",
                "max_distance": f"Max. distance in kilometers: {max_distance}",
                "primary": primary_name,
                "secondary": secondary_name,
            }
        )
        metadata["distance"] = xr.DataArray(
            distances, dims=("collocation",),
            attrs={
                "max_interval": f"Max. interval in secs: {max_interval}",
                "max_distance": f"Max. distance in kilometers: {max_distance}",
                "primary": primary_name,
                "secondary": secondary_name,
                "units": "kilometers",
            }
        )
        metadata["group"] = xr.DataArray(
            [primary_name, secondary_name], dims=("group",),
            attrs={
                "max_interval": f"Max. interval in secs: {max_interval}",
                "max_distance": f"Max. distance in kilometers: {max_distance}",
            }
        )

        output = add_xarray_groups(
            output, Collocations=metadata
        )

        start = pd.Timestamp(
            output[primary_name+"/time"].min().item(0)
        )
        end = pd.Timestamp(
            output[primary_name+"/time"].max().item(0)
        )

        output.attrs = {
            "start_time": str(start),
            "end_time": str(end),
        }

        return output

    @staticmethod
    def get_meta_group():
        return f"Collocations"

    def spatial_search_with_temporal_binning(
            self, primary, secondary, max_distance, max_interval
    ):
        # For time-binning purposes, pandas Dataframe objects are a good choice
        primary = pd.DataFrame(primary).set_index("time")
        secondary = pd.DataFrame(secondary).set_index("time")

        # Now let's split the two data data along their time coordinate so
        # we avoid searching for spatial collocations that do not fulfill
        # the temporal condition in the first place. However, the overhead
        # of the finding algorithm must be considered too (for example the
        # BallTree creation time). This can be adjusted by the parameter
        # bin_factor:
        bin_duration = self.bin_factor * max_interval

        # The binning is more efficient if we use the largest dataset as
        # primary:
        swapped_datasets = secondary.size > primary.size
        if swapped_datasets:
            primary, secondary = secondary, primary

        # Let's bin the primaries along their time axis and search for the
        # corresponding secondary bins:
        bin_pairs = (
            self._bin_pairs(start, chunk, primary, secondary, max_interval)
            for start, chunk in primary.groupby(pd.Grouper(freq=bin_duration))
        )

        # Add arguments to the bins (we need them for the spatial search
        # function):
        bins_with_args = (
            [self, max_distance, *bin_pair]
            for bin_pair in bin_pairs
        )

        # Unfortunately, a first attempt parallelizing this using threads
        # worsened the performance. Update: The BallTree code from scikit-learn
        # does not release the GIL. But apparently there will be a new version
        # coming that solves this problem, see this scikit-learn issue:
        # https://github.com/scikit-learn/scikit-learn/pull/10887. So stay
        # tuned!
        # threads = 1 if self.threads is None else self.threads
        t = Timer(verbose=False).start()
        # with ThreadPoolExecutor(max_workers=2) as pool:
        #     results = list(pool.map(
        #         Collocator._spatial_search_bin, bins_with_args
        #     ))

        results = list(map(
            Collocator._spatial_search_bin, bins_with_args
        ))

        self._debug(f"Collocated {len(results)} bins in {t.stop()}")

        pairs_list, distances_list = zip(*results)
        pairs = np.hstack(pairs_list)

        # No collocations were found.
        if not pairs.any():
            return self.no_pairs, self.no_distances

        # Stack the rest of the results together:
        distances = np.hstack(distances_list)

        if swapped_datasets:
            # Swap the rows of the results
            pairs[[0, 1]] = pairs[[1, 0]]

        return pairs.astype("int64"), distances

    @staticmethod
    def _bin_pairs(chunk1_start, chunk1, primary, secondary, max_interval):
        """"""
        chunk2_start = chunk1_start - max_interval
        chunk2_end = chunk1.index.max() + max_interval
        offset1 = primary.index.searchsorted(chunk1_start)
        offset2 = secondary.index.searchsorted(chunk2_start)
        chunk2 = secondary.loc[chunk2_start:chunk2_end]
        return offset1, chunk1, offset2, chunk2

    @staticmethod
    def _spatial_search_bin(args):
        self, max_distance, offset1, data1, offset2, data2 = args

        if data1.empty or data2.empty:
            return self.no_pairs, self.no_distances

        pairs, distances = self.spatial_search(
            data1["lat"].values, data1["lon"].values,
            data2["lat"].values, data2["lon"].values, max_distance
        )
        pairs[0] += offset1
        pairs[1] += offset2
        return pairs, distances

    def spatial_search(self, lat1, lon1, lat2, lon2, max_distance):
        # Finding collocations is expensive, therefore we want to optimize it
        # and have to decide which points to use for the index building.
        index_with_primary = self._choose_points_to_build_index(
            [lat1, lon1], [lat2, lon2],
        )

        self.index_with_primary = index_with_primary

        if index_with_primary:
            build_points = lat1, lon1
            query_points = lat2, lon2
        else:
            build_points = lat2, lon2
            query_points = lat1, lon1

        self.index = self._build_spatial_index(*build_points)
        pairs, distances = self.index.query(*query_points, r=max_distance)

        # No collocations were found.
        if not pairs.any():
            # We return empty arrays to have consistent return values:
            return self.no_pairs, self.no_distances

        if not index_with_primary:
            # The primary indices should be in the first row, the secondary
            # indices in the second:
            pairs[[0, 1]] = pairs[[1, 0]]

        return pairs, distances

    def _build_spatial_index(self, lat, lon):
        # Find out whether the cached index still works with the new points:
        if self._spatial_is_cached(lat, lon):
            self._debug("Spatial index is cached and can be reused")
            return self.index

        return GeoIndex(lat, lon, leaf_size=self.leaf_size)

    def _spatial_is_cached(self, lat, lon):
        """Return True if the cached ball tree is still applicable to the new
        data"""
        if self.index is None:
            return False

        try:
            return np.allclose(lat, self.index.lat) \
                   & np.allclose(lon, self.index.lon)
        except ValueError:
            # The shapes are different
            return False

    def _choose_points_to_build_index(self, primary, secondary):
        """Choose which points should be used for tree building

        This method helps to optimize the performance.

        Args:
            primary: Converted primary points
            secondary: Converted secondary points

        Returns:
            True if primary points should be used for tree building. False
            otherwise.
        """
        # There are two options to optimize the performance:
        # A) Cache the index and reuse it if either the primary or the
        # secondary points have not changed (that is the case for data with a
        # fixed grid). Building the tree is normally very expensive, so it
        # should never be done without a reason.
        # B) Build the tree with the larger set of points and query it with the
        # smaller set.
        # Which option should be used if A and B cannot be applied at the same
        # time? If the magnitude of one point set is much larger (by
        # `magnitude factor` larger) than the other point set, we strictly
        # follow B. Otherwise, we prioritize A.

        if primary[0].size > secondary[0].size * self.magnitude_factor:
            # Use primary points
            return True
        elif secondary[0].size > primary[0].size * self.magnitude_factor:
            # Use secondary points
            return False

        # Apparently, none of the datasets is much larger than the others. So
        # just check whether we still have a cached tree. If we used the
        # primary points last time and they still fit, use them again:
        if self.index_with_primary and self._spatial_is_cached(*primary):
            return True

        # Check the same for the secondary data:
        if not self.index_with_primary and self._spatial_is_cached(*secondary):
            return False

        # Otherwise, just use the larger dataset:
        return primary[0].size > secondary[0].size

    def temporal_search(self, primary, secondary, max_interval):
        raise NotImplementedError("Not yet implemented!")
        #return self.no_pairs, self.no_intervals

    def _temporal_check(
            self, primary_time, secondary_time, max_interval
    ):
        """Checks whether the current collocations fulfill temporal conditions

        Returns:

        """
        intervals = self._get_intervals(primary_time, secondary_time)

        # Check whether the time differences are less than the temporal
        # boundary:
        passed_time_check = intervals < max_interval

        return passed_time_check, intervals[passed_time_check]

    @staticmethod
    def _get_intervals(time1, time2):
        return np.abs((time1 - time2)).astype("timedelta64[s]")

    @staticmethod
    def _get_distances(lat1, lon1, lat2, lon2):
        return great_circle_distance(lat1, lon1, lat2, lon2)


def concat_collocations(collocations):
    """Concat compact collocations

    Compact collocations cannot be concatenated directly because indices in
    *Collocations/pairs* won't be correct any longer afterwards. This
    concatenate function fixes this problem.

    Args:
        collocations: A list of xarray.Dataset objects
            with compact collocations.

    Returns:
        One xarray.Dataset object
    """

    # We need to increment the pair indices when concatening the datasets
    primary = collocations[0]["Collocations/group"].item(0)
    secondary = collocations[0]["Collocations/group"].item(1)
    primary_size = 0
    secondary_size = 0
    collocation_coord = {
        "Collocations": "Collocations/collocation",
        primary: f"{primary}/collocation",
        secondary: f"{secondary}/collocation",
    }

    # Collect all collocations for each single group:
    groups = defaultdict(list)
    for obj in collocations:
        for group, data in get_xarray_groups(obj).items():
            if group == "Collocations":
                # Correct the indices:
                data["Collocations/pairs"][0, :] += primary_size
                data["Collocations/pairs"][1, :] += secondary_size
                data = data.drop("Collocations/group")
            groups[group].append(data)

        primary_size += obj.dims[f"{primary}/collocation"]
        secondary_size += obj.dims[f"{secondary}/collocation"]

    starts = []
    ends = []
    for group, data_list in groups.items():
        groups[group] = xr.concat(
                data_list,
                dim=collocation_coord[group]
            )

    start = pd.Timestamp(groups[primary][primary+"/time"].min().item(0))
    end = pd.Timestamp(groups[primary][primary+"/time"].max().item(0))
    merged = xr.merge(groups.values())
    merged.attrs = {
        "start_time": str(start),
        "end_time": str(end),
    }
    merged["Collocations/group"] = collocations[0]["Collocations/group"]
    return merged


class InvalidCollocationData(Exception):
    """Error when trying to collapse / expand invalid collocation data

    """

    def __init__(self, message, *args):
        Exception.__init__(self, message, *args)


def check_collocation_data(dataset):
    """Check whether the dataset fulfills the standard of collocated data

    Args:
        dataset: A xarray.Dataset object

    Raises:
        A InvalidCollocationData Error if the dataset did not pass the test.
    """
    mandatory_fields = ["Collocations/pairs", "Collocations/group"]

    for mandatory_field in mandatory_fields:
        if mandatory_field not in dataset.variables:
            raise InvalidCollocationData(
                f"Could not find the field '{mandatory_field}'!"
            )



# -*- coding: utf-8 -*-
"""Testing the basic geodetic functions.
"""
import numpy as np

from typhon import constants
from typhon import geodesy


class TestGeodesy:
    """Testing the geodetic functions."""
    def test_ellipsoidmodels(self):
        """Check ellipsoidmodels for valid excentricities."""
        e = geodesy.ellipsoidmodels()

        exc = np.array([e[m][1] for m in e.models])

        assert (np.all(exc >= 0) and np.all(exc < 1))

    def test_sind(self):
        """Test sinus calculation with degrees."""
        assert geodesy.sind(45) == np.sin(np.deg2rad(45))

    def test_cosd(self):
        """Test cosinus calculation with degrees."""
        assert geodesy.cosd(45) == np.cos(np.deg2rad(45))

    def test_tand(self):
        """Test tangens calculation with degrees."""
        assert geodesy.tand(45) == np.tan(np.deg2rad(45))

    def test_asind(self):
        """Test arcsinus calculation with degrees."""
        assert geodesy.asind(45) == np.arcsin(np.deg2rad(45))

    def test_cart2geocentric(self):
        """Test conversion from cartesian to geocentric system."""
        cartesian = (np.array([1, 0, 0]),  # x
                     np.array([0, 1, 0]),  # y
                     np.array([0, 0, 1]),  # z
                     )

        reference = (np.array([1, 1, 1]),  # r
                     np.array([0, 0, 90]),  # lat
                     np.array([0, 90, 0]),  # lon
                     )

        conversion = geodesy.cart2geocentric(*cartesian)

        assert np.allclose(conversion, reference)

    def test_geocentric2cart(self):
        """Test conversion from geocentric to cartesian system."""
        geocentric = (np.array([1, 1, 1]),  # r
                      np.array([0, 0, 90]),  # lat
                      np.array([0, 90, 0]),  # lon
                      )

        reference = (np.array([1, 0, 0]),  # x
                     np.array([0, 1, 0]),  # y
                     np.array([0, 0, 1]),  # z
                     )

        conversion = geodesy.geocentric2cart(*geocentric)

        assert np.allclose(conversion, reference)

    def test_geocentric2cart2geocentric(self):
        """Test conversion from geocentric to cartesian system and back."""
        ref = (1, -13, 42)

        cart = geodesy.geocentric2cart(*ref)
        geo = geodesy.cart2geocentric(*cart)

        assert np.allclose(ref, geo)

    def test_cart2geodetic(self):
        """Test conversion from cartesian to geodetic system."""
        r = geodesy.ellipsoidmodels().get('WGS84')[0]

        cartesian = (np.array([r, 0]),  # x
                     np.array([0, r]),  # y
                     np.array([0, 0]),  # z
                     )

        reference = (np.array([0, 0]),  # r
                     np.array([0, 00]),  # lat
                     np.array([0, 90]),  # lon
                     )

        conversion = geodesy.cart2geodetic(*cartesian)

        assert np.allclose(conversion, reference)

    def test_geodetic2cart(self):
        """Test conversion from geodetic to cartesian system."""
        r = geodesy.ellipsoidmodels().get('WGS84')[0]

        geodetic = (np.array([0, 0]),  # r
                    np.array([0, 0]),  # lat
                    np.array([0, 90]),  # lon
                    )

        reference = (np.array([r, 0]),  # x
                     np.array([0, r]),  # y
                     np.array([0, 0]),  # z
                     )

        conversion = geodesy.geodetic2cart(*geodetic)

        assert np.allclose(conversion, reference)

    def test_geodetic2cart2geodetic(self):
        """Test conversion from geodetic to cartesian system and back."""
        ellipsoid = geodesy.ellipsoidmodels().get('WGS84')
        ref = (1, -13, 42)

        cart = geodesy.geodetic2cart(*ref, ellipsoid)
        geod = geodesy.cart2geodetic(*cart, ellipsoid)

        assert np.allclose(ref, geod)

    def test_geodetic2geocentric2geodetic(self):
        """Test conversion from geodetic to geocentric system and back."""
        ellipsoid = geodesy.ellipsoidmodels().get('WGS84')
        ref = (1, -13, 42)

        geoc = geodesy.geodetic2geocentric(*ref, ellipsoid)
        geod = geodesy.geocentric2geodetic(*geoc, ellipsoid)

        assert np.allclose(ref, geod)

    def test_ellipsoid_r_geodetic(self):
        """Test return of geodetic radius for all ellipsois."""
        ellipsoid = geodesy.ellipsoidmodels().get('WGS84')

        r = geodesy.ellipsoid_r_geodetic(ellipsoid, 0)

        # Radius at equator has to be equal to the one defined in the
        # ellipsoidmodel.
        assert ellipsoid[0] == r

    def test_ellipsoid_r_geocentric(self):
        """Test return of geocentric radius for all ellipsois."""
        ellipsoid = geodesy.ellipsoidmodels().get('WGS84')

        r = geodesy.ellipsoid_r_geocentric(ellipsoid, 0)

        # Radius at equator has to be equal to the one defined in the
        # ellipsoidmodel.
        assert ellipsoid[0] == r

    def test_ellipsoid2d(self):
        """Test the calculation of new inclinated ellipsoids."""
        ellipsoid = geodesy.ellipsoidmodels().get('WGS84')

        e = geodesy.ellipsoid2d(ellipsoid, 0)

        assert e == (ellipsoid[0], 0)

    def test_ellipsoidcurvradius(self):
        """Test the calculation of local curvature radius."""
        ellipsoid = geodesy.ellipsoidmodels().get('WGS84')

        e = geodesy.ellipsoidcurvradius(ellipsoid, 0, 90)

        assert e == (ellipsoid[0], 0)

    def test_great_circle_distance(self):
        """Test calculation of great circle distance."""
        distance = geodesy.great_circle_distance(90, 30, 100, 60)

        assert np.isclose(distance, 10)

    def test_great_circle_distance_radius(self):
        """Test calculation of great circle distance with radious."""
        r = constants.earth_radius

        distance = geodesy.great_circle_distance(0, 0, 180, 0, r=r)

        assert distance == np.pi * r

    def test_geogeraphic_mean(self):
        """Test calculation of geographical mean of coordinates."""
        mean = geodesy.geographic_mean(0, [89, 90, 91])

        assert mean == (0, 90)

    def test_cartposlos2geocentric(self):
        """Test conversion of cartesian POS/LOS to spherical coordinates."""
        # TODO: Consider to implement a test for intuitive values.
        # The current reference are just arbitrary chosen to check that the
        # behaviour in future implementations does not change.

        # Reference svn revision 10041.
        reference = (
            np.array([1.73205081, 1.8493242]),
            np.array([35.26438968, 36.49922243]),
            np.array([45., 42.27368901]),
            np.array([0., 2.53049304]),
            np.array([0., 118.40177722])
            )

        results = geodesy.cartposlos2geocentric(
            [1, 1.1], [1, 1], [1, 1.1], [1, 1], [1, 1], [1, 1])

        assert np.allclose(results, reference)

    def test_cartposlos2geocentric_scalar(self):
        """Test conversion of cartesian POS/LOS for scalar input."""
        # The output is casted into ndarrays even if all input arguments are
        # floats.
        reference = (
            np.array([1.73205081]),
            np.array([35.26438968]),
            np.array([45.]),
            np.array([0.]),
            np.array([0.])
            )

        results = geodesy.cartposlos2geocentric(1, 1, 1, 1, 1, 1)

        assert np.allclose(results, reference)



"""Functions concerning the Optimal Estimation Method (OEM). """

from scipy.linalg import inv


__all__ = [
    'error_covariance_matrix',
    'averaging_kernel_matrix',
    'retrieval_gain_matrix',
]


def error_covariance_matrix(K, S_a, S_y):
    """Calculate the error covariance matrix.

    Parameters:
        K (np.array): Simulated Jacobians.
        S_a (np.array): A priori error covariance matrix.
        S_y (np.array): Measurement covariance matrix.

    Returns:
        np.array: Measurement error covariance matrix.
    """
    return inv(K.T @ inv(S_y) @ K + inv(S_a))


def averaging_kernel_matrix(K, S_a, S_y):
    """Calculate the averaging kernel matrix.

    Parameters:
        K (np.array): Simulated Jacobians.
        S_a (np.array): A priori error covariance matrix.
        S_y (np.array): Measurement covariance matrix.

    Returns:
        np.array: Averaging kernel matrix.
    """
    return retrieval_gain_matrix(K, S_a, S_y) @ K


def retrieval_gain_matrix(K, S_a, S_y):
    """Calculate the averaging kernel matrix.

    Parameters:
        K (np.array): Simulated Jacobians.
        S_a (np.array): A priori error covariance matrix.
        S_y (np.array): Measurement covariance matrix.

    Returns:
        np.array: Retrieval gain matrix.
    """
    return inv(inv(S_a) + K.T @ inv(S_y) @ K) @ K.T @ inv(S_y)



from os.path import dirname, join

import datetime
import numpy as np
import pytest

from typhon.files import FileHandler, FileInfo, FileSet, FileSetManager
from typhon.files.utils import get_testfiles_directory


class TestFileSet:
    """Testing the fileset methods."""

    filesets = None
    refdir = get_testfiles_directory("filesets")

    def init_filesets(self):
        if self.filesets is not None:
            return self.filesets

        self.filesets = FileSetManager()

        self.filesets += FileSet(
            join(
                self.refdir,
                "tutorial", "{satellite}", "{year}-{month}-{day}",
                "{hour}{minute}{second}-{end_hour}{end_minute}{end_second}.nc"
            ),
            name="tutorial",
        )

        self.filesets += FileSet(
            join(self.refdir, "single_file.nc",),
            name="single",
            time_coverage=["2018-01-01", "2018-01-03"],
        )

        def sequence_get_info(file_info, **kwargs):
            """Small helper function for sequence fileset."""
            with open(file_info) as f:
                file_info.times[0] = datetime.datetime.strptime(
                    f.readline().rstrip(),
                    "Start: %Y-%m-%d %H:%M:%S"
                )
                file_info.times[1] = datetime.datetime.strptime(
                    f.readline().rstrip(),
                    "End: %Y-%m-%d %H:%M:%S"
                )
            return file_info

        self.filesets += FileSet(
            join(self.refdir, "sequence",
                 "{year}", "{doy}", "sequence*.txt",),
            name="sequence-wildcard",
            handler=FileHandler(
                info=sequence_get_info,
            ),
            info_via="handler",
        )
        self.filesets += FileSet(
            join(self.refdir, "sequence",
                 "{year}", "{doy}", "sequence{id}.txt",
            ),
            handler=FileHandler(
                info=sequence_get_info,
            ),
            name="sequence-placeholder",
            info_via="both",
            placeholder={"id": r"\d{4}"}
        )

        self.filesets += FileSet(
            join(self.refdir,
                 # NSS.HIRX.NJ.D99127.S0632.E0820.B2241718.WI.gz
                 "regex", "NSS.HIR[XS].{satcode}.D{year2}{doy}.S{hour}"
                 "{minute}.E{end_hour}{end_minute}.B{B}.{station}.gz"
            ),
            name="regex-HIRS",
        )
        self.filesets["regex-HIRS"].set_placeholders(
            satcode=".{2}", B=r"\d{7}", station=".{2}",
        )

        return self.filesets

    @pytest.mark.skipif(refdir is None, reason="typhon-testfiles not found.")
    def test_contains(self):
        """Test whether all filesets cover the testing timestamps.

        Returns:
            None
        """
        filesets = self.init_filesets()
        tests = [
            # [Timestamp(s), Should it be covered by the filesets?]
            ["2016-01-01", False],
            ["2018-01-01", True],
            ["2018-01-01 06:00:00", True],
            [datetime.datetime(2018, 1, 1), True],
            [datetime.datetime(2018, 1, 1, 12,), True],
        ]

        for name, fileset in filesets.items():
            # print("Run test-contains for %s fileset:" % name)

            for timestamp, check in tests:
                # print("\tCheck coverage of %s (expected %s, got %s)" % (
                #     timestamp, check, timestamp in fileset
                # ))
                assert (timestamp in fileset) == check

    @pytest.mark.skipif(refdir is None, reason="typhon-testfiles not found.")
    def test_glob(self):
        files = FileSet(
            join(
                self.refdir,
                "tutorial", "{satellite}", "*", "*.nc"
            ),
            placeholder={"satellite": 'SatelliteA'},
        )

        # Sort this after paths rather than times (because the times are all
        # equal)
        check = list(sorted([
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteA', '2018-01-02', '000000-040000.nc'),
                [datetime.datetime(1, 1, 1, 0, 0),
                 datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)],
                {'satellite': 'SatelliteA'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteA', '2018-01-02', '080000-120000.nc'),
                [datetime.datetime(1, 1, 1, 0, 0),
                 datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)],
                {'satellite': 'SatelliteA'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteA', '2018-01-02', '200000-000000.nc'),
                [datetime.datetime(1, 1, 1, 0, 0),
                 datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)],
                {'satellite': 'SatelliteA'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteA', '2018-01-02', '040000-080000.nc'),
                [datetime.datetime(1, 1, 1, 0, 0),
                 datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)],
                {'satellite': 'SatelliteA'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteA', '2018-01-02', '120000-160000.nc'),
                [datetime.datetime(1, 1, 1, 0, 0),
                 datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)],
                {'satellite': 'SatelliteA'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteA', '2018-01-02', '160000-200000.nc'),
                [datetime.datetime(1, 1, 1, 0, 0),
                 datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)],
                {'satellite': 'SatelliteA'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteA', '2018-01-01', '000000-040000.nc'),
                [datetime.datetime(1, 1, 1, 0, 0),
                 datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)],
                {'satellite': 'SatelliteA'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteA', '2018-01-01', '080000-120000.nc'),
                [datetime.datetime(1, 1, 1, 0, 0),
                 datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)],
                {'satellite': 'SatelliteA'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteA', '2018-01-01', '200000-000000.nc'),
                [datetime.datetime(1, 1, 1, 0, 0),
                 datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)],
                {'satellite': 'SatelliteA'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteA', '2018-01-01', '040000-080000.nc'),
                [datetime.datetime(1, 1, 1, 0, 0),
                 datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)],
                {'satellite': 'SatelliteA'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteA', '2018-01-01', '120000-160000.nc'),
                [datetime.datetime(1, 1, 1, 0, 0),
                 datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)],
                {'satellite': 'SatelliteA'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteA', '2018-01-01', '160000-200000.nc'),
                [datetime.datetime(1, 1, 1, 0, 0),
                 datetime.datetime(9999, 12, 31, 23, 59, 59, 999999)],
                {'satellite': 'SatelliteA'}),
        ], key=lambda x: x.path))

        assert list(sorted(files, key=lambda x: x.path)) == check

    @pytest.mark.skip
    @pytest.mark.skipif(refdir is None, reason="typhon-testfiles not found.")
    def test_magic_methods(self):
        """Test magic methods on the fileset examples of the tutorial.

        Returns:

        """
        filesets = self.init_filesets()
        filters = {"satellite": "SatelliteB"}
        result = filesets["tutorial"]["2018-01-01 03:00", filters]

        check = np.array(
            ['2018-01-01T00:00:00.000000000', '2018-01-01T00:20:00.000000000',
             '2018-01-01T00:40:00.000000000', '2018-01-01T01:00:00.000000000',
             '2018-01-01T01:20:00.000000000', '2018-01-01T01:40:00.000000000',
             '2018-01-01T02:00:00.000000000', '2018-01-01T02:20:00.000000000',
             '2018-01-01T02:40:00.000000000', '2018-01-01T03:00:00.000000000',
             '2018-01-01T03:20:00.000000000', '2018-01-01T03:40:00.000000000',
             '2018-01-01T04:00:00.000000000', '2018-01-01T04:20:00.000000000',
             '2018-01-01T04:40:00.000000000'], dtype='datetime64[ns]')

        assert np.allclose(
            result["time"].astype("int"),
            check.astype("int")
        )

    @pytest.mark.skipif(refdir is None, reason="typhon-testfiles not found.")
    def test_tutorial(self):
        """Test the fileset examples of the tutorial.

        Returns:
            None
        """
        filesets = self.init_filesets()

        # STANDARD DATASET
        # Should not find anything:
        empty = list(
            filesets["tutorial"].find(
                "2017-12-31", "2018-01-01", no_files_error=False
            ))
        assert not empty

        # Find the closest file to 2018-01-01, limited to SatelliteB
        # temporarily:
        found_file = filesets["tutorial"].find_closest(
            "2018-01-01 03:00", filters={
                "!satellite": ("SatelliteA", "SatelliteC")
            }
        )

        #print("closest check", self._repr_file_info(found_file))

        check = FileInfo(
            join(self.refdir, 'tutorial',
                 'SatelliteB', '2018-01-01', '000000-050000.nc'),
            [datetime.datetime(2018, 1, 1, 0, 0),
             datetime.datetime(2018, 1, 1, 5, 0)], {'satellite': 'SatelliteB'})

        assert found_file == check

        # Limit this fileset to SatelliteB permanently
        filesets["tutorial"].set_placeholders(
            satellite="SatelliteB",
        )

        # Should find four files:
        found_files = list(
            filesets["tutorial"].find(
                "2018-01-01", "2018-01-02",
            ))

        #print("four files:")
        # self._print_files(found_files)

        check = [
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteB', '2018-01-01', '000000-050000.nc'),
                [datetime.datetime(2018, 1, 1, 0, 0),
                 datetime.datetime(2018, 1, 1, 5, 0)],
                {'satellite': 'SatelliteB'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteB', '2018-01-01', '050000-100000.nc'),
                [datetime.datetime(2018, 1, 1, 5, 0),
                 datetime.datetime(2018, 1, 1, 10, 0)],
                {'satellite': 'SatelliteB'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteB', '2018-01-01', '100000-150000.nc'),
                [datetime.datetime(2018, 1, 1, 10, 0),
                 datetime.datetime(2018, 1, 1, 15, 0)],
                {'satellite': 'SatelliteB'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteB', '2018-01-01', '150000-200000.nc'),
                [datetime.datetime(2018, 1, 1, 15, 0),
                 datetime.datetime(2018, 1, 1, 20, 0)],
                {'satellite': 'SatelliteB'}),
            FileInfo(
                join(self.refdir, 'tutorial',
                     'SatelliteB', '2018-01-01', '200000-010000.nc'),
                [datetime.datetime(2018, 1, 1, 20, 0),
                 datetime.datetime(2018, 1, 2, 1, 0)],
                {'satellite': 'SatelliteB'}),
        ]

        assert found_files == check

        # Should find four files and should return them in two bins:
        found_files = list(
            filesets["tutorial"].find(
                "2018-01-01", "2018-01-02", bundle="12h",
            ))

        # print("Bundle 12h:")
        # self._print_files(found_files)

        check = [
            [
                FileInfo(
                    join(self.refdir, 'tutorial',
                         'SatelliteB', '2018-01-01', '000000-050000.nc'),
                    [datetime.datetime(2018, 1, 1, 0, 0),
                     datetime.datetime(2018, 1, 1, 5, 0)],
                    {'satellite': 'SatelliteB'}),
                FileInfo(
                    join(self.refdir, 'tutorial',
                         'SatelliteB', '2018-01-01', '050000-100000.nc'),
                    [datetime.datetime(2018, 1, 1, 5, 0),
                     datetime.datetime(2018, 1, 1, 10, 0)],
                    {'satellite': 'SatelliteB'}),
                FileInfo(
                    join(self.refdir, 'tutorial',
                         'SatelliteB', '2018-01-01', '100000-150000.nc'),
                    [datetime.datetime(2018, 1, 1, 10, 0),
                     datetime.datetime(2018, 1, 1, 15, 0)],
                    {'satellite': 'SatelliteB'}),
            ],
            [
                FileInfo(
                    join(self.refdir, 'tutorial',
                         'SatelliteB', '2018-01-01', '150000-200000.nc'),
                    [datetime.datetime(2018, 1, 1, 15, 0),
                     datetime.datetime(2018, 1, 1, 20, 0)],
                    {'satellite': 'SatelliteB'}),
                FileInfo(
                    join(self.refdir, 'tutorial',
                         'SatelliteB', '2018-01-01', '200000-010000.nc'),
                    [datetime.datetime(2018, 1, 1, 20, 0),
                     datetime.datetime(2018, 1, 2, 1, 0)],
                    {'satellite': 'SatelliteB'}),
            ],
        ]

        assert found_files == check

        # Should find four files and should return them in two bins:
        found_files = list(
            filesets["tutorial"].find(
                "2018-01-01", "2018-01-02", bundle=3,
            ))

        # print("Bundle 3:")
        # self._print_files(found_files)

        check = [
            [
                FileInfo(
                    join(self.refdir, 'tutorial',
                         'SatelliteB', '2018-01-01', '000000-050000.nc'),
                    [datetime.datetime(2018, 1, 1, 0, 0),
                     datetime.datetime(2018, 1, 1, 5, 0)],
                    {'satellite': 'SatelliteB'}),
                FileInfo(
                    join(self.refdir, 'tutorial',
                         'SatelliteB', '2018-01-01', '050000-100000.nc'),
                    [datetime.datetime(2018, 1, 1, 5, 0),
                     datetime.datetime(2018, 1, 1, 10, 0)],
                    {'satellite': 'SatelliteB'}),
                FileInfo(
                    join(self.refdir, 'tutorial',
                         'SatelliteB', '2018-01-01', '100000-150000.nc'),
                    [datetime.datetime(2018, 1, 1, 10, 0),
                     datetime.datetime(2018, 1, 1, 15, 0)],
                    {'satellite': 'SatelliteB'}),
            ],
            [
                FileInfo(
                    join(self.refdir, 'tutorial',
                         'SatelliteB', '2018-01-01', '150000-200000.nc'),
                    [datetime.datetime(2018, 1, 1, 15, 0),
                     datetime.datetime(2018, 1, 1, 20, 0)],
                    {'satellite': 'SatelliteB'}),
                FileInfo(
                    join(self.refdir, 'tutorial',
                         'SatelliteB', '2018-01-01', '200000-010000.nc'),
                    [datetime.datetime(2018, 1, 1, 20, 0),
                     datetime.datetime(2018, 1, 2, 1, 0)],
                    {'satellite': 'SatelliteB'}),
            ],
        ]

        assert found_files == check

        for test_method in [FileSet.map, FileSet.imap]:
            # Check map method
            results = list(test_method(
                filesets["tutorial"], TestFileSet._tutorial_map,
                start="2018-01-01", end="2018-01-03"
            ))
            check = ['SatelliteB', 'SatelliteB', 'SatelliteB', 'SatelliteB',
                     'SatelliteB', 'SatelliteB', 'SatelliteB', 'SatelliteB',
                     'SatelliteB', 'SatelliteB']
            assert results == check

            # Check map method on content
            results = list(test_method(
                filesets["tutorial"], TestFileSet._tutorial_map_content,
                start="2018-01-01", end="2018-01-03", on_content=True,
            ))
            check = [111.92121062601221, 24.438060320121387,
                     -98.80775640366036, -75.84330354813459, 59.41297628327247,
                     106.80513550614192, -3.999061608822918,
                     -108.68523313569861, -51.82441769876156, 66.33842832792985
            ]
            assert np.allclose(results, check)

    @staticmethod
    def _tutorial_map(file_info):
        return file_info.attr["satellite"]

    @staticmethod
    def _tutorial_map_content(data,):
        return data["data"].mean().item(0)

    @pytest.mark.skipif(refdir is None, reason="typhon-testfiles not found.")
    def test_files_overlap_subdirectory(self):
        """A file covers a time period longer than its sub directory.
        """
        filesets = self.init_filesets()
        filesets["tutorial"].set_placeholders(
            satellite="SatelliteA"
        )
        found_file = filesets["tutorial"].find_closest("2018-01-03")

        check = FileInfo(
            join(self.refdir, 'tutorial',
                 'SatelliteA', '2018-01-02', '200000-000000.nc'),
            [datetime.datetime(2018, 1, 2, 20, 0),
             datetime.datetime(2018, 1, 3, 0, 0)],
            {'satellite': 'SatelliteA'}
        )

        assert found_file == check

    @pytest.mark.skipif(refdir is None, reason="typhon-testfiles not found.")
    def test_single(self):
        """Test find on the single fileset.

        Returns:
            None
        """
        filesets = self.init_filesets()

        # STANDARD DATASET
        # Should not find anything:
        empty = list(
            filesets["single"].find(
                "2016-12-31", "2018-01-01", no_files_error=False
            ))
        assert not empty

        check = [
            FileInfo(join(self.refdir, 'single_file.nc'),
                     [datetime.datetime(2018, 1, 1, 0, 0),
                      datetime.datetime(2018, 1, 3, 0, 0)], {}),
        ]

        found_files = list(
            filesets["single"].find(
                "2018-01-01", "2018-01-02",
            ))

        assert found_files == check

        found_files = list(
            filesets["single"].find(
                "2018-01-01", "2018-01-02", bundle="12h",
            ))

        assert found_files == check

        found_files = list(
            filesets["single"].find(
                "2018-01-01", "2018-01-02", bundle=3,
            ))

        assert found_files == check

    @pytest.mark.skipif(refdir is None, reason="typhon-testfiles not found.")
    def test_sequence(self):
        """Test find on the sequence filesets.

        Returns:
            None
        """
        filesets = self.init_filesets()

        # STANDARD DATASET
        # Should not find anything:
        empty = list(
            filesets["sequence-placeholder"].find(
                "2016-12-31", "2018-01-01", no_files_error=False
            ))
        assert not empty

        # Should find two files:
        found_files = list(
            filesets["sequence-placeholder"].find(
                "2018-01-01", "2018-01-02",
            ))

        check = [
            FileInfo(join(self.refdir, 'sequence',
                          '2018', '001', 'sequence0001.txt'),
                     [datetime.datetime(2018, 1, 1, 0, 0),
                      datetime.datetime(2018, 1, 1, 12, 0)], {'id': 1}),
            FileInfo(join(self.refdir, 'sequence',
                          '2018', '001', 'sequence0002.txt'),
                     [datetime.datetime(2018, 1, 1, 12, 0),
                      datetime.datetime(2018, 1, 2, 0, 0)], {'id': 2}),

        ]
        assert found_files == check

        # Should find two files and should return them in two bins:
        found_files = list(
            filesets["sequence-placeholder"].find(
                "2018-01-01", "2018-01-02", bundle="6h",
            ))

        check = [
            [
                FileInfo(join(self.refdir, 'sequence',
                              '2018', '001', 'sequence0001.txt'),
                         [datetime.datetime(2018, 1, 1, 0, 0),
                          datetime.datetime(2018, 1, 1, 12, 0)], {'id': 1}),
            ],
            [
                FileInfo(join(self.refdir, 'sequence',
                              '2018', '001', 'sequence0002.txt'),
                         [datetime.datetime(2018, 1, 1, 12, 0),
                          datetime.datetime(2018, 1, 2, 0, 0)], {'id': 2}),
            ],
        ]
        assert found_files == check

    @pytest.mark.skipif(refdir is None, reason="typhon-testfiles not found.")
    def test_sequence_placeholder(self):
        """Test find on all standard filesets.

        Returns:
            None
        """
        filesets = self.init_filesets()

        # STANDARD DATASET
        # Should not find anything:
        empty = list(
            filesets["sequence-placeholder"].find(
                "2016-12-31", "2018-01-01", no_files_error=False
            ))
        assert not empty

        # Should find two files:
        found_files = list(
            filesets["sequence-placeholder"].find(
                "2018-01-01", "2018-01-02",
            ))

        check = [
            FileInfo(join(self.refdir, 'sequence',
                          '2018', '001', 'sequence0001.txt'),
                     [datetime.datetime(2018, 1, 1, 0, 0),
                      datetime.datetime(2018, 1, 1, 12, 0)], {'id': 1}),
            FileInfo(join(self.refdir, 'sequence',
                          '2018', '001', 'sequence0002.txt'),
                     [datetime.datetime(2018, 1, 1, 12, 0),
                      datetime.datetime(2018, 1, 2, 0, 0)], {'id': 2}),

        ]
        assert found_files == check

        # Should find two files and should return them in two bins:
        found_files = list(
            filesets["sequence-placeholder"].find(
                "2018-01-01", "2018-01-02", bundle="6h",
            ))

        check = [
            [
                FileInfo(join(self.refdir, 'sequence',
                              '2018', '001', 'sequence0001.txt'),
                         [datetime.datetime(2018, 1, 1, 0, 0),
                          datetime.datetime(2018, 1, 1, 12, 0)], {'id': 1}),
            ],
            [
                FileInfo(join(self.refdir, 'sequence',
                              '2018', '001', 'sequence0002.txt'),
                         [datetime.datetime(2018, 1, 1, 12, 0),
                          datetime.datetime(2018, 1, 2, 0, 0)], {'id': 2}),
            ],
        ]
        assert found_files == check

    @pytest.mark.skipif(refdir is None, reason="typhon-testfiles not found.")
    def test_regex(self):
        filesets = self.init_filesets()

        check = [
            FileInfo(join(self.refdir, 'regex',
                          'NSS.HIRX.NJ.D99127.S0632.E0820.B2241718.WI.gz'),
                     [datetime.datetime(1999, 5, 7, 6, 32),
                      datetime.datetime(1999, 5, 7, 8, 20)],
                     {'satcode': 'NJ', 'B': '2241718',
                      'station': 'WI'}),
        ]

        found_file = filesets["regex-HIRS"].find_closest("1999-05-08")

        assert found_file == check[0]
        assert found_file.attr == check[0].attr

        found_files = \
            list(filesets["regex-HIRS"].find("1999-05-07", "1999-05-09"))

        assert found_files == check

    @pytest.mark.skipif(refdir is None, reason="typhon-testfiles not found.")
    def test_complicated_subdirs(self, ):
        """Check whether FileSet can find files in subdirectories that contain
        text and placeholders.
        """
        # The Pinocchio fileset from the cloud toolbox: a folder name contains
        # normal text and a placeholder:
        pinocchio = FileSet(
            join(self.refdir, "pinocchio",
                 "t{year2}{month}{day}",
                 "tm{year2}{month}{day}{hour}{minute}{second}{millisecond}.jpg",
                 ),
        )

        # Find all files:
        files = list(pinocchio)

        check = [
            FileInfo(
                join(self.refdir,
                     'pinocchio', 't171102', 'tm171102132855573.jpg'),
                [datetime.datetime(2017, 11, 2, 13, 28, 55, 573000),
                 datetime.datetime(2017, 11, 2, 13, 28, 55, 573000)], {}),
        ]
        assert files == check

    @pytest.mark.skip
    def test_align(self):
        """Test the align method.
        """
        # Find an adequate test

        filesets = self.init_filesets()

        start, end = "2018-01-01", "2018-01-02"
        a_dataset = filesets["tutorial"].copy()
        a_dataset.name = "SatelliteA"
        a_dataset.set_placeholders(satellite="SatelliteA")
        a_reference = a_dataset.collect(start, end)
        b_dataset = filesets["tutorial"].copy()
        b_dataset.name = "SatelliteB"
        b_dataset.set_placeholders(satellite="SatelliteB")
        b_reference = b_dataset.collect(start, end)

        all_data = {}
        for files, data in a_dataset.align(start, end, [b_dataset]):
            for dataset in data:
                if dataset in all_data:
                    all_data[dataset] = GroupedArrays.concat(
                        [all_data[dataset], *data[dataset]]
                    )
                else:
                    all_data[dataset] = GroupedArrays.concat(
                        data[dataset]
                    )

        a_retrieved = all_data["SatelliteA"]
        b_retrieved = all_data["SatelliteB"]

        assert a_reference == a_retrieved
        assert b_reference == b_retrieved

    def _repr_file_info(self, file_info):

        path = "join(self.refdir, '%s')" % (
            file_info.path[len(self.refdir)+1:]
        )

        return "FileInfo(\n\t{}, \t{}, \t{}),".format(
            path, repr(file_info.times), repr(file_info.attr)
        )



"""Datasets for models
"""

# Any commits made to this module between 2015-05-01 and 2017-03-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822
# 
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

import datetime

import numpy

from . import dataset

class ERAInterim(dataset.NetCDFDataset, dataset.MultiFileDataset):
    # example path:
    # /badc/ecmwf-era-interim/data/gg/as/2015/01/01/ggas201501011200.nc
    name = section = "era_interim"
    subdir = "data/{AA:s}/{B:s}{C:s}/{year:04d}/{month:02d}/{day:02d}"
    re = (r"(?P<AA>[a-z]{2})(?P<B>[a-z])(?P<C>[a-z])"
          r"(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})"
          r"(?P<hour>\d{2})(?P<minute>\d{2})\.nc")
    start_date = datetime.datetime(1979, 1, 1, 0, 0, 0)
    end_date = datetime.datetime(2016, 1, 1, 0, 0, 0)
    granule_duration = datetime.timedelta(hours=6)
    def _read(self, f, *args, **kwargs):
        # may need to convert, but I currently need gg/as which on CEMS is
        # already in converted format, so will assume NetCDF
        (M, extra) = super()._read(f, *args,
                          pseudo_fields={"time": self._get_time_from_ds},
                          prim="t",
                          **kwargs)
        return (M, extra)

    @staticmethod
    def _get_time_from_ds(ds):
        epoch = datetime.datetime.strptime(ds["t"].time_origin,
                                           "%d-%b-%Y:%H:%M:%S")
        return numpy.datetime64(epoch) + ds["t"][:].astype("timedelta64[D]")



# -*- coding: utf-8 -*-
"""Testing the basic nonlte functions.
"""
import numpy as np

from typhon import nonlte


class TestNonlte:
    """Testing the nonlte functions."""
    def test_trapz_inte_edge(self):
        """Check trapzouidal integration function."""
        x = np.arange(1, 11, 1)
        y = np.random.random(10)

        area = nonlte.mathmatics.trapz_inte_edge(y, x)
        area_ref = np.trapz(y, x)
        assert np.allclose(area.sum(), area_ref)



# -*- coding: utf-8 -*-

"""This module contains convenience functions for general file handling.
"""

from .fileset import *
from .handlers import *
from .utils import *

__all__ = [s for s in dir() if not s.startswith('_')]


r"""
An implementation of quantile regression neural networks based on the
`Keras <https://github.com/keras-team/keras>` deep learning package.
The implementation has been developed specifically for remote sensing
applications and provides a high level interface allowing for simple
training and evaluation of the QRNNs."""
from typhon.retrieval.qrnn.qrnn import QRNN



# -*- coding: utf-8 -*-
"""Testing the functions in typhon.latex.
"""
import os
from tempfile import mkstemp

import numpy as np

from typhon import latex


class TestLaTeX:
    """Testing the latex functions."""
    ref_dir = os.path.join(os.path.dirname(__file__), "reference", "")

    def setup_method(self):
        """Create a temporary file."""
        fd, self.f = mkstemp()
        os.close(fd)

    def teardown_method(self):
        """Delete temporary file."""
        os.remove(self.f)

    def test_texify_matrix(self):
        """Save NumPy array as LaTeX table."""
        latex.texify_matrix(
            np.arange(20).reshape(5, 4),
            fmt="%.3f",
            filename=self.f,
            caption="This is a test caption.",
            heading=['H1', 'H2', 'H3', 'H4'],
            align='c',
            delimiter=False
            )

        ref = os.path.join(self.ref_dir, 'matrix.tex')
        with open(self.f) as testfile, open(ref) as reffile:
            assert testfile.readlines() == reffile.readlines()



"""
This module provides classes and functions to find collocations between
filesets and datasets. They are inspired by the CollocatedDataset classes in
atmlab implemented by Gerrit Holl.

Created by John Mrziglod, June 2017
"""

import logging

import numba
import numpy as np
from typhon.files import FileSet
from typhon.utils.timeutils import Timer
import xarray as xr

from .collocator import check_collocation_data, Collocator

__all__ = [
    "collapse",
    "Collocator",
    "Collocations",
    "expand",
]

logger = logging.getLogger(__name__)


class Collocations(FileSet):
    """Class for finding and storing collocations between FileSet objects

    If you want to find collocations between Arrays, use :func:`collocate`
    instead.
    """
    def __init__(
            self, *args, reference=None, read_mode=None, collapser=None,
            **kwargs):
        """Initialize a Collocation object

        This :class:`~typhon.files.fileset.FileSet`

        Args:
            *args: Positional arguments for
                :class:`~typhon.files.fileset.FileSet`.
            read_mode: The collocations can be collapsed or expanded after
                collecting. Set this either to *collapse* (default),
                *expand* or *compact*.
            reference: If `read_mode` is *collapse*, here you can set the name
                of the dataset to that the others should be collapsed. Default
                is the primary dataset.
            collapser: If `read_mode` is *collapse*, here you can give your
                dictionary with additional collapser functions.
            **kwargs: Keyword arguments for
                :class:`~typhon.files.fileset.FileSet`.
        """
        # Call the base class initializer
        super().__init__(*args, **kwargs)

        self.read_mode = read_mode
        self.reference = reference
        self.collapser = collapser
        self.collocator = None

    def add_fields(self, original_fileset, fields, **kwargs):
        """

        Args:
            start:
            end:
            original_fileset:
            group
            fields:

        Returns:
            None
        """
        map_args = {
            "on_content": True,
            "kwargs": {
                "original_fileset": original_fileset,
                "fields": fields,
            },
            **kwargs,
        }

        return self.map(Collocations._add_fields, **map_args)

    @staticmethod
    def _add_fields(data, original_fileset, fields):
        pass

    def read(self, *args, **kwargs):
        """Read a file and apply a collapse / expand function to it

        Does the same as :meth:`~typhon.files.fileset.FileSet.read` from the
        base class :class:`~typhon.files.fileset.FileSet`, but can
        collapse or expand collocations after reading them.

        Args:
            *args: Positional arguments for
                :meth:`~typhon.files.fileset.FileSet.read`.
            **kwargs: Keyword arguments for
                :meth:`~typhon.files.fileset.FileSet.read`.

        Returns:
            The same as :meth:`~typhon.files.fileset.FileSet.read`, but with
            data that is either collapsed or expanded.
        """
        data = super(Collocations, self).read(*args, **kwargs)

        if self.read_mode == "compact":
            # Do nothing
            return data
        elif self.read_mode == "collapse" or self.read_mode is None:
            # Collapse the data (default)
            return collapse(data, self.reference, self.collapser)
        elif self.read_mode == "expand":
            # Expand the data
            return expand(data)
        else:
            raise ValueError(
                f"Unknown reading read_mode for collocations: "
                f"{self.read_mode}!\nAllowed read_modes are: 'collapse' "
                f"(default), 'expand' or 'compact'."
            )

    def search(self, filesets, collocator=None, **kwargs):
        """Find all collocations between two filesets and store them in files

        Collocations are two or more data points that are located close to each
        other in space and/or time.

        This takes all files from the filesets between two dates, find
        collocations of their data points and store them in output files.

        Each collocation output file provides these standard fields:

        * *fileset_name/lat* - Latitudes of the collocations.
        * *fileset_name/lon* - Longitude of the collocations.
        * *fileset_name/time* - Timestamp of the collocations.
        * *fileset_name/__index* - Indices of the collocation data in
            the original files.
        * *fileset_name/__file_start* - Start time of the original file.
        * *fileset_name/__file_end* - End time of the original file.
        * *collocations/pairs* - Tells you which data points are collocated
            with each other by giving their indices.

        And all additional fields that were loaded from the original files (you
        can control which fields should be loaded by the `read_args` parameter
        from the :class:`~typhon.files.fileset.FileSet` objects in `filesets`).
        Note that subgroups in the original data will be flattened by replacing
        */* with *_* since xarray is not yet able to handle grouped data
        properly.

        Args:
            filesets: A list of FileSet objects.
            collocator: If you want to use your own collocator, you can pass it
                here. It must behave like
                :class:`~typhon.collocations.collocator.Collocator`.
            **kwargs: Additional keyword arguments that are allowed for
                :meth:`~typhon.collocations.collocator.Collocator.collocate_filesets`.

        Examples:

        .. :code-block:: python

            from typhon.collocations import Collocations
            from typhon.files import FileSet, MHS_HDF, SEVIRI


            # Define a fileset with the files from MHS / NOAA18:
            mhs = FileSet(
                name="MHS",
                path="/path/to/files/{year}/{month}/{day}/*NSS.MHSX.*."
                     "S{hour}{minute}.E{end_hour}{end_minute}.*.h5",
                handler=MHS_HDF(),
                read_args={
                    "fields": ["Data/btemps"],
                }
            )

            # Define a fileset with files from SEVIRI
            seviri = FileSet(
                name="SEVIRI",
                path="/path/{year}/{month}/{day}/{hour}{minute}{second}.*.h5.gz",  # noqa
                handler=SEVIRI(),
                # Each file of CloudSat covers exactly 15 minutes.
                time_coverage="15 minutes",
            )

            # Define where the collocations should be stored:
            collocations = Collocations(
                path="/path/MHS-{satname}_SEVIRI/{year}/{month}/{day}/"
                     "{hour}{minute}{second}-{end_hour}{end_minute}{end_second}.nc"
            )

            # Search for collocations and store them
            collocations.search(
                [seviri, mhs], start="2013-12-10", end="20 Dec 2013",
                processes=10, max_interval="5 min", max_distance="5 km",
            )
        """
        if collocator is None:
            collocator = Collocator()

        collocated_files = collocator.collocate_filesets(
            filesets, output=self, **kwargs
        )

        timer = Timer().start()
        for _ in collocated_files:
            pass
        logger.info(f"{timer} for finding all collocations")


@numba.jit
def _rows_for_secondaries_numba(primary):
    """Helper function for collapse - numba optimized"""
    current_row = np.zeros(primary.size, dtype=int)
    rows = np.zeros(primary.size, dtype=int)
    i = 0
    for p in primary:
        rows[i] = current_row[p]
        i += 1
        current_row[p] += 1
    return rows


def _rows_for_secondaries(primary):
    """Helper function for collapse"""
    current_row = np.zeros(primary.size, dtype=int)
    rows = np.zeros(primary.size, dtype=int)
    i = 0
    for p in primary:
        rows[i] = current_row[p]
        i += 1
        current_row[p] += 1
    return rows


def collapse(data, reference=None, collapser=None):
    """Collapse all multiple collocation points to a single data point

    During searching for collocations, one might find multiple collocation
    points from one dataset for one single point of the other dataset. For
    example, the MHS instrument has a larger footprint than the AVHRR
    instrument, hence one will find several AVHRR colloocation points for
    each MHS data point. This method performs a function on the multiple
    collocation points to merge them to one single point (e.g. the mean
    function).

    Args:
        data:
        reference: Normally the name of the dataset with the largest
            footprints. All other dataset will be collapsed to its data points.
        collapser: Dictionary with names of collapser functions to apply and
            references to them. Defaults collpaser functions are *mean*, *std*
            and *number* (count of valid data points).

    Returns:
        A xr.Dataset object with the collapsed data

    Examples:
        .. code-block:: python

            # TODO: Add examples
    """
    collapsed = xr.Dataset(
        attrs=data.attrs.copy(),
    )

    # Check whether the collocation data is compatible
    check_collocation_data(data)

    pairs = data["Collocations/pairs"].values
    groups = data["Collocations/group"].values.tolist()

    if reference is None:
        # Take automatically the first member of the groups (was the primary
        # while collocating)
        reference = groups[0]

    if reference not in groups:
        raise ValueError(
            f"The selected reference '{reference}' is not valid, because it "
            f"was not collocated! Collocated groups were: {groups}."
        )

    # Find the index of the reference group. If it is the primary, it is 0
    # otherwise 1.
    reference_index = groups[0] != reference

    primary_indices = pairs[int(reference_index)]
    secondary_indices = pairs[int(not reference_index)]

    # THE GOAL: We want to bin the secondary data according to the
    # primary indices and apply a collapse function (e.g. mean) to it.
    # THE PROBLEM: We might to group the data in many (!) bins that might
    # not have the same size and we have to apply a function onto each of
    # these bins. How to make this efficient?
    # APPROACH 1: pandas and xarray provide the powerful groupby method
    # which allows grouping by bins and applying a function to it.
    # -> This does not scale well with big datasets (100k of primaries
    # takes ~15 seconds).
    # APPROACH 2: Applying functions onto matrix columns is very fast in
    # numpy even if the matrix is big. We could create a matrix where each
    # column acts as a primary bin, the number of its rows is the maximum
    # number of secondaries per bin. Then, we fill the secondary data into
    # the corresponding bins. Since they might be a different number of
    # secondaries for each bin, there will be unfilled slots. We fill these
    # slots with NaN values (so they won't affect the outcome of the
    # collapser function). Now, we can apply the collapser function on the
    # whole matrix along the columns.
    # -> This approach is very fast but might need more space.
    # We follow approach 2, since it may scale better than approach 1 and
    # we normally do not have to worry about memory space. Gerrit's
    # collocation toolkit in atmlab also follows a similar approach.
    # Ok, let's start!

    # The matrix has the shape of N(max. number of secondaries per primary)
    # x N(unique primaries). So the columns are the primary bins. We know at
    # which column to put the secondary data by using primary_indices. Now, we
    # have to find out at which row to put them. For big datasets, this might
    # be a very expensive function. Therefore, we have have two version: One
    # (pure-python), which we use for small datasets, and one (numba-optimized)
    # for big datasets because Numba produces an overhead:
    #timer = time.time()
    if len(primary_indices) < 1000:
        rows_in_bins = _rows_for_secondaries(primary_indices)
    #    print(f"{time.time()-timer:.2f} seconds for pure-python")
    else:
        rows_in_bins = _rows_for_secondaries_numba(primary_indices)
    #    print(f"{time.time()-timer:.2f} seconds for numba")

    # The user may give his own collapser functions:
    if collapser is None:
        collapser = {}
    collapser = {
        "mean": lambda m, a: np.nanmean(m, axis=a),
        "std": lambda m, a: np.nanstd(m, axis=a),
        "number": lambda m, a: np.count_nonzero(~np.isnan(m), axis=a),
        **collapser,
    }

    for var_name, var_data in data.variables.items():
        group, local_name = var_name.split("/", 1)

        # We copy the reference, collapse the rest and ignore the meta data of
        # the collocations (because they become useless now)
        if group == "Collocations":
            continue

        # This is the name of the dimension along which we collapse:
        collapse_dim = group + "/collocation"

        if collapse_dim not in var_data.dims:
            # This variable does not depend on the collocation coordinate.
            # Hence, we cannot collapse it but we simply copy it. To make the
            # dataset collocation-friendly, the time, lat and lon fields of the
            # reference group will be copied to the root path:
            if group == reference and local_name in ("time", "lat", "lon"):
                collapsed[local_name] = var_data
            else:
                collapsed[var_name] = var_data

            continue

        # Make sure that our collapsing dimension is the first dimension of the
        # array. Otherwise we get problems, when converting the DataArray to a
        # numpy array.
        new_ordered_dims = list(var_data.dims)
        new_ordered_dims.remove(collapse_dim)
        new_ordered_dims.insert(0, collapse_dim)
        var_data = var_data.transpose(*new_ordered_dims)

        # Rename the first dimension to the main collocation dimension
        dims = list(var_data.dims)
        dims[0] = "collocation"
        var_data.dims = dims

        if group == reference:
            # We want to make the resulting dataset collocation-friendly (so
            # that we might use it for a collocation search with another
            # dataset). So the content of the reference group moves "upward"
            # (group name vanishes from the path):
            if local_name in ("time", "lat", "lon"):
                collapsed[local_name] = var_data
            else:
                collapsed[var_name] = var_data
            continue

        # The standard fields (time, lat, lon) and the special fields to
        # retrieve additional fields are useless after collapsing. Hence,
        # ignore them (won't be copied to the resulting dataset):
        if local_name in ("time", "lat", "lon") or local_name.startswith("__"):
            continue

        # The dimensions of the binner matrix:
        binner_dims = [
            np.max(rows_in_bins) + 1,
            np.unique(primary_indices).size
        ]

        # The data might have additional dimensions (e.g. brightness
        # temperatures from MHS have 5 channels). We must now how many
        # additional dimensions the data have and what their length is.
        add_dim_sizes = [
            var_data.shape[i]
            for i, dim in enumerate(var_data.dims)
            if dim != "collocation"
        ]
        binner_dims.extend(add_dim_sizes)

        # Fill the data in the bins:
        # Create an empty matrix:
        binned_data = np.empty(binner_dims)

        # Fill all slots with NaNs:
        binned_data[:] = np.nan
        binned_data[rows_in_bins, primary_indices] \
            = var_data.isel(collocation=secondary_indices).values

        for func_name, func in collapser.items():
            collapsed[f"{var_name}_{func_name}"] = \
                var_data.dims, func(binned_data, 0)

    return collapsed


def expand(dataset):
    """Repeat the primary data so they align with their secondary collocations

    During searching for collocations, one might find multiple collocation
    points from one dataset for one single point of the other dataset. For
    example, the MHS instrument has a larger footprint than the AVHRR
    instrument, hence one will find several AVHRR colloocation points for
    each MHS data point. To avoid needing more storage than required, no
    duplicated data values are stored even if they collocate multiple times.

    Args:
        dataset: A xarray.Dataset object with collocated data.

    Returns:
        A xarray.Dataset object.
    """
    # Check whether the collocation data is compatible
    check_collocation_data(dataset)

    pairs = dataset["Collocations/pairs"].values
    groups = dataset["Collocations/group"].values.tolist()

    expanded = dataset.isel(
        **{groups[0] + "/collocation": pairs[0]}
    )
    expanded = expanded.isel(
        **{groups[1] + "/collocation": pairs[1]}
    )

    # The collocation coordinate of all datasets are equal now:
    for i in range(2):
        expanded["collocation"] = groups[i] + "/collocation", \
                                  np.arange(pairs[i].size)
        expanded = expanded.swap_dims(
            {groups[i] + "/collocation": "collocation"}
        )

    # The variable pairs is useless now:
    expanded = expanded.drop("Collocations/pairs")

    expanded = expanded.rename({"Collocations/collocation": "collocation"})

    return expanded



"""Various statistical functions

"""

# Any commits made to this module between 2015-05-01 and 2017-03-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822
# 
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

import numbers

import numpy

import scipy.stats
import scipy.special


def bin(x, y, bins):
    """Bin/bucket y according to values of x.

    Returns list of arrays, one element with values for bin.
    Digitising happens with numpy.digitize.

    You can use this for geographical data, but if your x and
    y are in lat/lons the bins will be equirectangular (if you
    want equal-area bins you will have to reproject / change
    coordinates).

    Arguments:

        x (ndarray): Coordinate that y is binned along
        y (ndarray):  Data to be binned.  First dimension must match
            length of x.  All subsequent dimensions are left untouched.
        bins (ndarray): Bins according to which sort data.

    Returns:

        List of arrays, one element per bin.
    """
    if x.size == y.size == 0:
        return [y[()] for b in bins]
    digits = numpy.digitize(x, bins)
    binned = [y[digits == i, ...] for i in range(len(bins))]
    return binned


def bin_nd(binners, bins, data=None):
    """Bin/bucket data in arbitrary number of dimensions

    For example, one can bin geographical data according to lat/lon
    through:

    >>> binned = bin_nd([lats, lons], [lat_bins, lon_bins])

    The actually binned data are the indices for the arrays lats/lons,
    which hopefully corresponds to indices in your actual data.

    Data that does not fit in any bin, is not binned anywhere.

    You can use this for geographical data, but if your x and
    y are in lat/lons the bins will be equirectangular (if you
    want equal-area bins you will have to reproject / change
    coordinates).

    Note: do NOT pass the 3rd argument, `data`.  This is used purely for
    the implementation using recursion.  Passing anything here explicitly
    is a recipe for disaster.

    Arguments:

        binners (List[ndarray]): Axes that data is binned at.  This is
            akin to the x-coordinate in `:func:bin`.

        bins (List[ndarray]): Edges for the bins according to which bin
            data.

    Returns:
        n-D ndarray of type `object`, with indices describing what bin
        elements belong to.
    """

    if len(bins) != len(binners):
        raise ValueError("Length of bins must equal length of binners. "
                         "Found {} bins, {} binners.".format(
                             len(bins), len(binners)))

    for b in bins:
        if b.ndim != 1:
            raise ValueError("Bin-array must be 1-D. "
                             "Found {}-D array.".format(b.ndim))

    if not all([b.size == binners[0].size for b in binners[1:]]):
        raise ValueError("All binners must have same length.")

    dims = numpy.array([b.size for b in bins])

    nd = len(binners)

    if nd == 0:
        return numpy.array([], dtype=numpy.uint64)

    indices = numpy.arange(binners[0].size, dtype=numpy.uint64)
    if data is None:
        data = indices

    if nd > 1:
        # innerbinned = bin(binners[-1], data, bins[-1])
        innerbinned = bin(binners[-1], indices, bins[-1])
        outerbinned = []
        for (i, ib) in enumerate(innerbinned):
            obinners = [x[ib] for x in binners[:-1]]
            ob = bin_nd(obinners, bins[:-1], data[ib])
            outerbinned.append(ob)

        # go through some effort to make sure v[i, j, ...] is always
        # numpy.uint64, whereas v is numpy.object_
        # do this in steps, see comment in the else:-block below for
        # reasoning
        #
        # We have outerbinned, which has length n_N, and contains ndarrays
        # of size n_1 * n_2 * ... * n_{N-1}.
        #
        # We want V to be n_1 * n_2 * ... * n_N, where N is the number of
        # dimensions we are binning.
        #
        # The following could /probably/ be do with some sophisticated
        # list comprehension and permutation, but this is clearer.

        V = numpy.empty(shape=dims, dtype=numpy.object_)
        for i in range(len(outerbinned)):
            V[..., i] = outerbinned[i]

#        V.T[...] = [
#            [numpy.array(e.tolist(), dtype=numpy.uint64)
#                for e in l] for l in outerbinned]
        return V
        # return numpy.array(v, dtype=numpy.object_)
    else:
        # NB: I should not convert a list-of-ndarrays to an object-ndarray
        # directly.  If all nd-arrays have the same dimensions (such as
        # size x=0), the converted nd-array will have x as an additional
        # dimension, rather than having object arrays inside the
        # container.  To prevent this, explicitly initialise the ndarray.
        binned = bin(binners[0], data, bins[0])
        B = numpy.empty(shape=len(binned), dtype=numpy.object_)
        B[:] = binned
        return B


def binned_statistic(coords, data, bins, statistic=None):
    """Bin data and calculate statistics on them

    As :func:`scipy.stats.binned_statistic` but faster for simple
    statistics.

    Args:
        coords: 1-dimensional numpy.array with the coordinates that should be
            binned.
        data: A numpy.array on which the statistic should be applied.
        bins: The bins used for the binning.
        statistic: So far only *mean* is implemented.

    Returns:
        A numpy array with statistic results.
    """
    bin_indices = numpy.digitize(coords, bins)
    data_sum = numpy.bincount(bin_indices, weights=data, minlength=bins.size)
    data_number = numpy.bincount(bin_indices, minlength=bins.size)
    return data_sum / data_number

def get_distribution_as_percentiles(x, y,
                                    bins,
                                    ptiles=(5, 25, 5, 75, 95)):
    """get the distribution of y vs. x as percentiles.

    Bin y-data according to x-data (using :func:`typhon.math.stats.bin`).
    Then, within each bin, calculate percentiles.

    Arguments:

        x (ndarray): data for x-axis
        y (ndarray): data for y-axis
        bins (ndarray): Specific bins to use for dividing the x-data.
        ptiles (ndarray): Percentiles to use.
    """

    # explicitly get rid of masked data, because scoreatpercentile is not
    # masked-array aware
    try:
        good = (~x.mask) & (~y.mask)
    except AttributeError: # not a masked-array, leave as-is
        pass
    else: # surely masked arrays
        x = x[good].data
        y = y[good].data
    ybinned = bin(x, y, bins)
    return numpy.vstack([scipy.stats.scoreatpercentile(b, ptiles)
                         for b in ybinned])


def adev(x, dim=-1):
    r"""Calculate Allan deviation in its simplest form

    Arguments:

        x (ndarray or xarray DataArray): n-dim array for Allan calculation
        dim (int or str): optional, axis to operate along, defaults to
            last.  If you pass a str, x must be a xarray.DataArray and the
            dimension will be a name.

    .. math::
        \sigma = \sqrt{\frac{1}{2(N-1)} \sum_{i=1}^{N-1} (y_{i+1} - y_i)^2}

    Equation source: Jon Mittaz, personal communication, April 2016
    """

    
    if isinstance(dim, numbers.Integral):
        # dimension by number, probably ndarray
        x = x.swapaxes(-1, dim)
        N = x.shape[-1]
        return numpy.sqrt(1/(2*(N-1)) *
                          ((x[..., 1:] - x[..., :-1])**2).sum(-1))
    else:
        # dimension by name, should be xarray.Dataarray
        N = x.sizes[dim]
        return numpy.sqrt((1/(2*(N-1)) * x.diff(dim=dim)**2).sum(dim=dim))

def corrcoef(mat):
    """Calculate correlation coefficient with p-values

    Calculate correlation coefficients along with p-values.

    Arguments:

        mat (ndarray): 2-D array [p×N] for which the correlation matrix is
            calculated

    Returns:

        (r, p) where r is a p×p matrix with the correlation coefficients,
        obtained with numpy.corrcoef, and p is 

    Attribution:
    
        this code, or an earlier version was posted by user 'jingchao' on
        Stack Overflow at 2014-7-3 at
        http://stackoverflow.com/a/24547964/974555 and is licensed under
        CC BY-SA 3.0.  This notice may not be removed.
    """
    r = numpy.corrcoef(mat)
    rf = r[numpy.triu_indices(r.shape[0], 1)]
    df = mat.shape[1] - 2
    ts = rf * rf * (df / (1 - rf * rf))
    pf = scipy.special.betainc(0.5 * df, 0.5, df / (df + ts))
    p = numpy.zeros(shape=r.shape)
    p[numpy.triu_indices(p.shape[0], 1)] = pf
    p[numpy.tril_indices(p.shape[0], -1)] = pf
    p[numpy.diag_indices(p.shape[0])] = numpy.ones(p.shape[0])
    return (r, p)
